import urllib

def connection_class_from_name(name):
    pass

def connection_from_string(connection_string):
    parsed = urllib.parse(connection_string)
    pass
    # Look at the scheme to decide what connection type to use.



#!/usr/bin/python3

from acomms import Micromodem, UnifiedLog, Message
from acomms.modem_connections import UdpConnection as ModemUdpConnection
from acomms.flexibledataprotocol import FDPacket, FDFrame
import argparse
import socket
from threading import Thread
from queue import Queue, Empty
from time import sleep
import logging

class UdpConnection(object):
    def __init__(self, receive_callback, remote_host, remote_port, local_host=None, local_port=None, timeout=0.1):
        self._incoming_line_buffer = ""

        self.receive_callback = receive_callback
        self.timeout = timeout
        self._remote_host = remote_host
        self._remote_port = remote_port

        if local_host is None:
            self._local_host = ""
        else:
            self._local_host = local_host

        if local_port is None:
            self._local_port = remote_port
        else:
            self._local_port = local_port

        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.bind(('', self._local_port))
        self._socket.settimeout(0.1)

        self._thread = Thread(target=self._listen)
        self._thread.setDaemon(True)
        self._thread.start()

    def close(self):
        self._thread.stop()

    def _listen(self):
        while True:
            msg_lines = self.readlines()
            # If we get data, we are connected, so pass through to the parser
            if msg_lines:
                for line in msg_lines:
                    self.receive_callback(line)

    def readlines(self):
        """Returns a \n terminated line  Only returns complete lines (or None on timeout)"""
        try:
            rl = self._socket.recv(1024)
        except socket.timeout:
            return None

        if len(rl) == 0:
            return None

        rl = rl.decode('ascii')

        self._incoming_line_buffer += rl

        # Make sure we got a complete line.  (We always should with RECON.)
        if '\n' in self._incoming_line_buffer:
            # is there a newline at the end?
            lines = self._incoming_line_buffer.splitlines(True)
            # See if the last line has a newline at the end.
            if lines[-1][-1] != '\n':
                self._incoming_line_buffer = lines[-1]
                lines.pop()  # remove it from the list to passed on
            else:
                self._incoming_line_buffer = ''
            # return the list of complete lines
            return lines
        else:
            return None

    def write(self, data):
        if type(data) is str:
            data = data.encode('ascii')
        self._socket.sendto(data, (self._remote_host, self._remote_port))

class FdpTranslator(object):
    def __init__(self, remote_address, remote_port, local_port=None, modem_serial_port=None, modem_baud=19200,
                 modem_remote_address=None, modem_remote_port=4001, modem_local_port=4001):
        self.modem_rx_queue = Queue()
        self.modem = Micromodem()
        # Now that we have a good time, start a new log with a good timestamp.
        unified_log = UnifiedLog(console_log_level=logging.INFO)
        self.modem.unified_log = unified_log
        if modem_serial_port:
            self.modem.connect_serial(modem_serial_port, modem_baud)
        else:
            self.modem.connection = ModemUdpConnection(self.modem, modem_remote_address,
                                                       modem_remote_port, local_port=modem_local_port)
        self.modem.attach_incoming_msg_queue(self.modem_rx_queue)

        self.udp_connection = UdpConnection(self.udp_receive_callback, remote_address, remote_port, local_port=local_port)

        self.modem_rx_thread = Thread(target=self.modem_receive_callback, daemon=True)
        self.modem_rx_thread.start()

    def modem_receive_callback(self):
        while True:
            try:
                message = self.modem_rx_queue.get(timeout=0.1)
                # Pass all messages through
                self.udp_connection.write(message['raw'].encode('ascii'))
                print("modem> {}".format(message["raw"].strip()))
                # CARFP messages will be handled by the modem directly
                # CAACK:
                if message['type'] == "CACDR":
                    # $CACDR,isodate,rxseqnum,src,dest,minirate,miniframe_ACK,dataframe_ACK,reserved1,…reserved6*CS
                    if int(message['params'][6]) > 0:
                        caack_msg = "$CAACK,{src},{dest},{frame_num},1\r\n".format(src=message['params'][2],
                                                                                   dest=message['params'][3],
                                                                                   frame_num=1)
                        self.udp_connection.write(caack_msg.encode('ascii'))
            except Empty:
                pass


    def udp_receive_callback(self, line):
        try:
            # Drop certain messages
            if line.startswith("$CCCYC"):
                cyc_msg = Message(line)
                self.udp_connection.write("$CACYC,{},{},{},{},{},{}\r\n".format(*cyc_msg['params']))
                self.udp_connection.write("$CADRQ,000000,{},{},0,32,1\r\n".format(cyc_msg['params'][1],
                                                                              cyc_msg['params'][2]))
            elif line.startswith("$CCTXD"):
                # turn this into a CCTFP message with ACK set
                txd_msg = Message(line) # $CCTXD,src,dest,ack,hexdata
                cctfp_msg = "$CCTFP,{dest},1,1,1,0,,{hexdata}".format(dest=txd_msg['params'][1],
                                                                      hexdata=txd_msg['params'][3])
                self.modem.write_nmea(cctfp_msg)
                print("modem< {}".format(cctfp_msg))
            # just pass through other messages
            else:
                self.modem.write_nmea(line)
        except Exception as e:
            print(e)

if __name__ == '__main__':
    ap = argparse.ArgumentParser(description='Connect to a MM for testing purposes')
    ap.add_argument("-C", "--COM", help='COM Port to connect', default="/dev/ttyO1")
    ap.add_argument("-BR", "--Baudrate", help="COM Port Baud Rate", default=19200)

    args = ap.parse_args()

    #translator = FdpTranslator("COM20", 19200, '10.11.62.211', 4001)
    translator = FdpTranslator(remote_address='10.19.30.199', remote_port=4002, local_port=4002,
                               modem_remote_address='10.19.30.61', modem_remote_port=4001, modem_local_port=4001)

    try:
        while True:
            sleep(1)
    finally:
        translator = None


class Ping(dict):

    fields = ('type',
              'src', 'dest',
              'rate',
              'isodate',
              'snr_in', 'snr_out',
              'owtt',
              'tx_level',
              'reserved1', 'reserved2',
              'timestamp_res', 'toa_mode',
              'snv_on', 'timestamp'
               )


    def __getattr__(self, item):
        """Maps values to attributes.
        Only called if there *isn't* an attribute with this name
        """
        try:
            return self.__getitem__(item)
        except KeyError:
            raise AttributeError(item)

    def __setattr__(self, item, value):
        """Maps attributes to values.
        Only if we are initialised
        """
        if self.__dict__.__has_key(item):
            dict.__setattr__(self, item, value)
        else:
            self.__setitem__(item, value)

    def __str__(self):
        hrstr = "{ts}\tSRC: {src: .0f}\tDST: {dst: .0f}\tOWTT: {owtt: .0f}\tSNR_out: {snr_out: .0f}\tTX_Level: {txlvel: .0f}".format(
            ts = self['isodate'], src = self['src'], dst = self['dest'], owtt = self['owtt'], snr_out = self['snr_out'], txlvel = self['tx_level'])

        return hrstr

    @classmethod
    def from_nmea_msg(cls, msg, modem):
        print(msg['type'])

        if not modem.fw_at_least("v3.0.0"):
            if msg['type'] == "CACMD":
                return cls(Ping.from_cacmd_msg(cls, msg, modem))
            elif msg['type'] == "CACMA":
                return cls(Ping.from_cacma_msg(cls, msg, modem))
        elif msg['type'] == "CACMA":
                return cls(Ping.from_cacma_msg(cls, msg, modem))
        elif msg['type'] == "CACMR":
                return cls(Ping.from_cacmr_msg(cls, msg, modem))

        modem._daemon_log.error("Ping: unrecognized nmea message.")
        return None

    def from_cacmd_msg(cls, msg, modem):
        values = dict.fromkeys(Ping.fields)

        if not modem.fw_at_least("v3.0.0"):
            if msg["params"][0] == "PNG":
                values['type'] = msg['params'][0]
                values['src'] = int(msg["params"][1])
                values['dest'] = int(msg["params"][2])
                values['rate'] = int(msg["params"][3])
                values['cdr'] = int(msg["params"][4])
            elif msg["params"][1] == "PNR":
                values['type'] = msg['params'][1]
                values['src'] = int(msg["params"][2])
                values['dest'] = int(msg["params"][3])
                values['owtt'] = float(msg["params"][4])
                values['snr_in'] = float(msg["params"][5])
                values['snr_out'] = float(msg["params"][6])
                values['tx_level'] = int(msg["params"][7])
                values['reserved1'] = int(msg["params"][8])
                values['reserved2'] = int(msg["params"][8])

            return values
        else:
            modem._daemon_log.error("Modem acknowledges ping command requested. Shouldn't try to parse")
            return None

    def from_cacma_msg(cls, msg, modem):
        values = dict.fromkeys(Ping.fields)
        if not modem.fw_at_least("v3.0.0"):
            values['isodate'] = msg['params'][0]
            values['type'] = msg['params'][1]

        else:
            values['type'] = msg['params'][0]
            values['isodate'] = msg['params'][1]

        values['src'] = int(msg['params'][2])
        values['dest'] = int(msg['params'][3])
        values['snr_in'] = float(msg['params'][4])
        values['snr_out'] = float(msg['params'][5])
        values['tx_lvel'] = float(msg['params'][6])
        values['reserved1'] = msg['params'][7]
        values['reserved2'] = msg['params'][8]

        return values

    def from_cacmr_msg(cls, msg, modem):
        values = dict.fromkeys(Ping.fields)

        values['type'] = msg['params'][0]
        values['isodate'] = msg['params'][1]
        values['src'] = int(msg["params"][2])
        values['dest'] = int(msg["params"][3])
        values['owtt'] = float(msg["params"][4])
        values['snr_in'] = float(msg["params"][5])
        values['snr_out'] = float(msg["params"][6])
        values['tx_level'] = int(msg["params"][7])
        values['reserved1'] = int(msg["params"][8])
        time_fields = (msg["params"][9]).split(';')
        if len(time_fields) == 4:
            values['timestamp_res'] = int(time_fields[0])
            values['toa_mode'] = int(time_fields[1])
            values['snv_on'] = int(time_fields[2])
            values['timestamp'] = int(time_fields[3])

        return values

    def is_reply(self):
        if self['type'] == 'PNR':
            return True
        else:
            return False



from acomms.flexibledataprotocol import FDPacket, FDFrame
from acomms.pid_defs import *
from acomms.index_list import IndexList

class NetPacket(object):

    def __init__(self, link_src = None, link_dest = None,
                 header_valid=False, header_bytes=None,
                 valid_index: IndexList = None, data_bytes=None):

        self.link_src = link_src
        self.link_dest = link_dest

        self.header_valid = header_valid

        if data_bytes is None and valid_index is not None:
            raise ValueError("valid_index requires data_bytes")

        if header_bytes is not None:
            self.header_bytes = header_bytes
        else:
            self.header_bytes = []

        if valid_index is not None:
            self.valid_index = valid_index
        elif valid_index is None and data_bytes is not None:
            self.valid_index = IndexList([0, len(data_bytes)])

        if data_bytes is not None:
            self.data_bytes = data_bytes
        else:
            self.data_bytes = []

        self.subheaders = {}

        if self.header_valid:
            self.subheaders, idx = self._split_header(self.header_bytes)

    @classmethod
    def from_fdpacket(cls, fdpacket: FDPacket):
        netpacket = cls()

        netpacket._init_header(fdpacket)
        netpacket._init_data(fdpacket)

        if netpacket.header_valid:
            netpacket.subheaders, idx = netpacket._split_header(netpacket.header_bytes)

        netpacket.link_src = fdpacket.src
        netpacket.link_dest = fdpacket.dest

        return netpacket

    def to_fdpacket(self, miniframerate=1, dataframerate=1, ack=False):
        fdpacket = FDPacket(self.link_src, self.link_dest,
                            miniframerate, dataframerate, ack,
                            minibytes=self.header_bytes, databytes=self.data_bytes)

        return fdpacket

    def _init_header(self, fdpacket: FDPacket):
        #if not fdpacket.miniframes:
            #self._init_header_data_legacy(fdpacket)
        #else:

        valid_index, header = self._combine_frames(fdpacket.miniframes)

        # If there are missing bytes in the header, we can't really use it
        if len(valid_index) != 1:
            self.header_valid = False
        # Also be sure the index covers full length of header
        elif valid_index[0][-1] != len(header):
            self.header_valid = False
        else:
            self.header_bytes = header
            self.header_valid = True

    def _init_data(self, fdpacket: FDPacket):
        # If we can't parse the header, nothing to do with this data
        if not self.header_valid:
            return

        # If fdpacket.miniframes is non populated we already parsed
        # header and data together in legacy configuration
        if not fdpacket.miniframes:
            return

        valid_index, data = self._combine_frames(fdpacket.dataframes)

        self.data_bytes = data
        self.valid_index = valid_index

    def _init_header_data_legacy(self, fdpacket):
        # If the first dataframe failed, we can't parse the header.
        if not fdpacket.dataframes[0].crccheck:
            self.header_valid = False
        else:
            valid_index, mixed_bytes = self._combine_frames(fdpacket.dataframes)

            subheaders, idx = self._split_header(mixed_bytes)

            if subheaders is not None:
                idx = idx+1
                self.header_bytes = mixed_bytes[0:idx]
                self.data_bytes = mixed_bytes[idx:]

                # Modify valid index for new data_byte index
                for i in range(len(valid_index)):
                    for k in range(len(valid_index[i])):
                        valid_index[i][k] = valid_index[i][k] - idx
                        if valid_index[i][k] < 0:
                            valid_index[i][k] = 0

                # Remove elements in index that are out of bounds
                new_index = []
                for i in range(len(valid_index)):
                    if valid_index[i][-1] > 0:
                        new_index.append(valid_index[i])

    def _split_header(self, headerbytes, valid_index = None):
        subheaders = dict()
        idx = 0
        while idx < len(headerbytes):
            # Check if we've exceeded the valid index for these bytes
            if valid_index is not None:
                if idx > valid_index[0][-1]:
                    return None, None
            # If we reached terminator PID, no more subheaders
            if headerbytes[idx] == END_PID:
                break
            # If the PID is in vetted address space, we should know the length
            if headerbytes[idx] < FIRST_PROTO_PID:
                pass  # TODO: PID length lookup from pid_defs.py
            # Else we are in prototype PID space, so use included length field
            else:
                subheaders[headerbytes[idx]] = headerbytes[idx: idx + headerbytes[idx + 1]]
                idx = idx + headerbytes[idx + 1]

        return subheaders, idx

    def _combine_frames(self, frames):
        valid_index = IndexList()
        fbytes = bytearray()

        idx = 0

        if len(frames) > 0:
            for frame in frames:
                length = frame.nbytes
                start = idx
                end = idx + length

                if frame.crccheck:
                    valid_index.append([start, end])
                    fbytes = fbytes + frame.data
                else:
                    fbytes = fbytes + bytearray(length)

                idx = idx + length

            #TODO: handle case where all frame CRC checks fail
            valid_index.simplify()

        return valid_index, fbytes



import argparse
from acomms import ccl


def do_decode(hex_data):
    ccl_msg = ccl.CclDecoder.decode_hex_string(hex_data)
    for (name, value) in ccl_msg.items():
        human_readable = '{}\t{}'.format(name, value).expandtabs(24)
        print(human_readable)


if __name__ == '__main__':
    ap = argparse.ArgumentParser(description='Decode a CCL message into human-readable things.')
    ap.add_argument("hex_data", help="Hex encoded data (RXD payload)")

    args = ap.parse_args()

    do_decode(args.hex_data)




"""Acoustic File-Transfer-Protocol Protocol-IDs"""
FTP_FILE_PID = 0xa0
#FTP_FILE_PID_LEN = 8

FTP_ACK_PID = 0xa1
#FTP_ACK_PID_LEN = 4
#FTP_ACK_BLOCK_LEN = 4

FTP_NET_PID = 0xa2
#FTP_NET_PID_LEN = 12

"""Begin prototype address space"""
FIRST_PROTO_PID = 0xa0
"""Header terminator PID"""
END_PID = 0xff



__author__ = 'Eric'

from bitstring import BitStream, BitArray
import bitstring
from datetime import datetime
from acomms.messageparams import data_from_hexstring
import math
from bidict import bidict

mission_modes = bidict({0: "Mission completed",
                        1: "Manual",
                        2: "Test",
                        3: "Fault",
                        4: "Home",
                        5: "Alt Mission",
                        6: "Mission",
                        })

ranger_mission_modes = {1: "Pre-mission",
                        2: "Run",
                        5: "Completed"}

# Objective names from VIP.  It looks like these sometimes change with VIP version.
objective_names = bidict({0: 'None',
                          1: 'Manual',
                          2: 'Bench Test',
                          3: 'Acoustic Test',
                          4: 'Hysteresis Test',
                          5: 'Fin Wiggle Test',
                          6: 'Abort',
                          7: 'Fin Step',
                          8: 'Wait Run',
                          11: 'Set Position',
                          12: 'Wait Depth',
                          13: 'Surface',
                          19: 'Wait Prop',
                          20: 'Compass Cal',
                          21: 'Navigate',
                          22: 'Navigate Rows',
                          23: 'Wait Magnet',
                          24: 'Loiter',
                          25: 'Get GPS Fix',
                          27: 'Drop Decent',
                          28: 'Drop Ascent',
                          29: 'RECON',
                          32: 'Reacquire',
                          37: 'Drive Decent',
                          39: 'Circle',
                          42: 'Dock Cable',
                          43: 'Find',
                          44: 'Stand Off',
                          47: 'INS Reset',
                          49: 'Circular Cal',
                          50: 'Autotune',
                          51: 'CCL C3',
                          53: 'Undock Cable',
                          54: 'Unstuck',
                          62: 'Dock Mooring',
                          63: 'Undock Mooring',
                          65: 'Surface Loiter',
                          66: 'Find Max',
                          68: 'Relative Loiter',
                          69: 'PHINS C7 DVL Cal',
                          })

command_codes = {1: "Abort to end",
                 2: "Abort immediately",
                 3: "Start mission",
                 7: "Enable ranger ping",
                 8: "Disable ranger ping",
                 15: "Dump redirect commands",
                 17: "Abort to start",
                 18: "Abort to destination",
                 19: "Dump redirect except current"}


def LatLon(object):
    def __init__(self, lat_degrees=None, lon_degrees=None, lat_minutes=0, lon_minutes=None, lat_seconds=0, lon_seconds=None):
        if lat_degrees is not None:
            self._lat_degrees = float(lat_degrees)
            self._lat_degrees += lat_minutes/60
            self._lat_degrees += lat_seconds/(60*60)

        if lon_degrees is not None:
            self._lon_degrees = float(lon_degrees)
            self._lon_degrees += lon_minutes/60
            self._lon_degrees += lon_seconds/(60*60)


def decode_latlon(bit_array):
    if bit_array is None:
        raise ValueError

    # switch to big endian
    bit_array.byteswap()

    # sign extend
    if bit_array[0] == 1:
        bit_array.prepend('0xFF')
    else:
        bit_array.prepend('0x00')

    # Do conversion and return
    return bit_array.intbe * (180.0 / (2**23 - 1))


def encode_latlon(lat_or_lon):
    encoded = lat_or_lon * ((2**23 - 1) / 180.0)

    s = bitstring.pack('int:24', encoded)
    s.byteswap()

    return s


def decode_time_date(databits, year=None):
    databits.byteswap()

    month = databits.read('uint:4')
    day = databits.read('uint:5')
    hour = databits.read('uint:5')
    minute = databits.read('uint:6')
    second = databits.read('uint:4') * 4

    if year == None:
        year = datetime.utcnow().year

    return datetime(year, month, day, hour, minute, second)


def encode_time_data(date_time):
    s = bitstring.pack('uint:4, uint:5, uint:5, uint:6, uint:4',
                       date_time.month, date_time.day,
                       date_time.hour, date_time.minute, (date_time.second // 4))
    s.byteswap()
    return s


def decode_depth(encoded_depth):
    if encoded_depth <= 1000:
        return encoded_depth * 0.1
    elif encoded_depth <= 1500:
        return 100 + (encoded_depth - 1000) * 0.2
    elif encoded_depth <= 3100:
        return 200 + (encoded_depth - 1500) * 0.5
    elif encoded_depth <= 8100:
        return 1000 + (encoded_depth - 3100)
    else:
        return 6000


def encode_depth(depth):
    """
    0 - 100 meters: 0 - 1000(10 cm resolution)
    100 - 200 meters: 1001 - 1500(20 cm resolution)
    200 - 1000 meters: 1501 - 3100(50 cm resolution)
    1000 - 6000 meters: 3101 - 8100(1 meter resolution)
    """
    if depth < 0:
        return 0
    elif depth < 100:
        return int((depth+0.05) / 0.1)
    elif depth < 200:
        return int((depth - 100 + 0.1) / 0.2 + 1000)
    elif depth < 1000:
        return int((depth - 200 + 0.25) / 0.5 + 1500)
    elif depth < 6000:
        return int((depth - 1000 + 0.5) + 3100)
    else:
        return 8100


def decode_mission_mode_and_depth(encoded_bits):
    encoded_bits.byteswap()

    mission_mode_code = encoded_bits[0:3].uint
    depth = decode_depth(encoded_bits[3:].uint)

    mission_mode = mission_modes[mission_mode_code]

    return (mission_mode, depth)


def encode_mission_mode_and_depth(mission_mode, depth):
    if not isinstance(mission_mode, int):
        mission_mode_code = mission_modes.inverse[mission_mode]
    else:
        mission_mode_code = mission_mode

    encoded_depth = encode_depth(depth)
    s = bitstring.pack('uint:3, uint:13', mission_mode_code, encoded_depth)
    s.byteswap()

    return s


def encode_gfi_oil_pitch(gfi_percent, oil, pitch):
    s = bitstring.pack('int:6, uint:5, uint:5',
                       pitch * 63.0 / 180,
                       oil * 31.0 / 100,
                       gfi_percent * 31.0 / 100)
    s.byteswap()
    return s


class CclDecoder(object):
    ''' This attaches to the incoming dataframe queue and raises events when recognized messages are received
    '''

    @staticmethod
    def decode_dataframe(dataframe):
        msg_class = CclTypes[dataframe.data[0]]
        ccl_message = msg_class.from_data(dataframe.data)
        ccl_message['src'] = dataframe.src
        ccl_message['dest'] = dataframe.dest

        return ccl_message

    @staticmethod
    def decode_data(data):
        msg_class = CclTypes[data[0]]
        ccl_message = msg_class.from_data(data)

        return ccl_message

    @staticmethod
    def decode_hex_string(hex_data):
        data = data_from_hexstring(hex_data)
        msg_class = CclTypes[data[0]]
        ccl_message = msg_class.from_data(data)

        return ccl_message

    def __init__(self, modem):
        modem.rxframe_listeners.append(self.on_modem_rxframe)
        self.modem = modem
        self.ccl_listeners = []

    def on_modem_rxframe(self, dataframe):
        # Check for all the CCL types
        # try:
        ccl_message = self.decode_dataframe(dataframe)
        self.on_ccl_received(ccl_message)


        #except:
        #    pass

    def on_ccl_received(self, ccl_message):
        for listener in self.ccl_listeners:
            listener(ccl_message)


class MdatCommandMessage(dict):
    fields = ('mode', 'command', 'command_code' 'parameter')

    def __init__(self, command_code=None, command=None):
        self['mode'] = 'MDAT_COMMAND'
        self['command_code'] = command_code
        self['command'] = command_codes[command_code]
        self['parameter'] = bytearray(28)

    @property
    def as_data(self):
        databytes = bytearray([11, 0, 0])
        databytes.append(self['command_code'])
        databytes.append(self['parameter'])
        return databytes

    @classmethod
    def from_data(cls, databytes):
        values = dict.fromkeys(MdatCommandMessage.fields)

        # accept either the full frame payload or just the data after the CCL type identifier
        if (len(databytes) > 31):
            databytes = databytes[1:-1]

        values['mode'] = 'MDAT_COMMAND'

        values['command_code'] = databytes[2]
        values['command'] = command_codes[databytes[2]]
        values['parameter'] = databytes[3:-1]

        # Make a message
        mdat_command = cls(values)

        return mdat_command


class MdatRangerMessage(dict):
    fields = ('mode', 'latitude', 'longitude',
              'fix_age', 'heading', 'mission_mode',
              'depth', 'major_fault', 'mission_leg', 'num_legs', 'speed_kt',
              'battery_percent')

    @classmethod
    def from_data(cls, databytes):
        values = dict.fromkeys(MdatRangerMessage.fields)

        # accept either the full frame payload or just the data after the CCL type identifier
        if (len(databytes) > 31):
            databytes = databytes[1:-1]

        values['mode'] = 'MDAT_RANGER'

        hex_string = ''.join('{:02x}'.format(x) for x in databytes)
        # Now, hex_string is a 62-character string

        # Latitude
        lat_deg = int(hex_string[1:3])
        lat_dir = 'N' if hex_string[3] == 'a' else 'S'
        lat_min = float(hex_string[4:10]) / 10000.
        values['latitude'] = "{}{}{}".format(lat_deg, lat_dir, lat_min)

        values['lat_deg'] = lat_deg
        values['lat_decmin'] = lat_min
        values['lat_dir'] = lat_dir


        # Longitude
        lon_deg = int(hex_string[10:13])
        lon_dir = 'E' if hex_string[13] == 'b' else 'W'
        lon_min = float(hex_string[14:20]) / 10000.

        values['longitude'] = "{}{}{}".format(lon_deg, lon_dir, lon_min)

        values['lon_deg'] = lon_deg
        values['lon_decmin'] = lon_min
        values['lon_dir'] = lon_dir



        values['mission_leg'] = int(hex_string[20:24])
        values['num_legs'] = int(hex_string[24:28])

        values['time_remaining'] = "{}:{}".format(hex_string[28:30], hex_string[30:32])

        values['battery_percent'] = int(hex_string[32:34])
        values['speed_kt'] = "{}.{}".format(hex_string[34], hex_string[35])

        values['heading'] = int(hex_string[36:39])

        values['fix_age'] = "{}:{}".format(hex_string[40:42], hex_string[42:44])

        values['depth'] = int(hex_string[44:46])

        values['mission_mode'] = ranger_mission_modes.get(int(hex_string[46:48]),
                                                          "Unknown mode ({})".format(hex_string[46:48]))

        values['major_fault'] = True if (hex_string[48] == '8') else False


        # Make a message
        mdat_ranger = cls(values)

        return mdat_ranger


class MdatStateMessage(dict):
    '''
    A single MDAT_STATE message
    '''

    # Note that we don't override the dict initalizer.

    fields = ('mode', 'latitude', 'longitude',
              'fix_age', 'time_date', 'heading', 'mission_mode',
              'depth', 'faults_bits', 'mission_leg', 'estimated_velocity', 'objective_index',
              'power_watts', 'goal_latitude', 'goal_longitude', 'battery_percent', 'gfi_percent',
              'pitch', 'oil')

    # This automagically retrieves values from the dictionary when they are referenced as properties.
    # Whether this is awesome or sucks is open to debate.
    def __getattr__(self, item):
        """Maps values to attributes.
        Only called if there *isn't* an attribute with this name
        """
        try:
            return self.__getitem__(item)
        except KeyError:
            raise AttributeError(item)

    def __setattr__(self, item, value):
        """Maps attributes to values.
        Only if we are initialised
        """
        if not self.__dict__.has_key(
                '_MdatStateMessage__initialized'):  # this test allows attributes to be set in the __init__ method
            return dict.__setattr__(self, item, value)
        elif self.__dict__.has_key(item):  # any normal attributes are handled normally
            dict.__setattr__(self, item, value)
        else:
            self.__setitem__(item, value)

    def __str__(self):
        ''' Default human-readable version
        Doesn't show all parameters, just the most common ones.'''
        hrstr = "State Message:\t{ts}\t{lat}\t{lon}".format(
            ts=self['time_date'], lat=self['latitude'], lon=self['longitude'])

        return hrstr


    @classmethod
    def from_data(cls, databytes):

        values = dict.fromkeys(MdatStateMessage.fields)

        # accept either the full frame payload or just the data after the CCL type identifier
        if (len(databytes) > 31):
            databytes = databytes[1:]

        d = BitStream(bytes=databytes)

        values['mode'] = 'MDAT_STATE'

        lat_bits = BitArray(d.read('bits:24'))
        values['latitude'] = decode_latlon(lat_bits)

        lon_bits = BitArray(d.read('bits:24'))
        values['longitude'] = decode_latlon(lon_bits)

        values['fix_age'] = d.read('uint:8') * 4
        values['time_date'] = decode_time_date(d.read('bits:24'))
        values['heading'] = d.read('uint:8') * (360.0 / 255.0)

        mission_mode_depth_bits = d.read('bits:16')
        (values['mission_mode'], values['depth']) = decode_mission_mode_and_depth(mission_mode_depth_bits)
        values['faults_bits'] = d.read('bits:40')
        values['mission_leg'] = d.read('uint:8')
        values['estimated_velocity'] = d.read('uint:8') / 25.0
        values['objective_index'] = d.read('uint:8')
        values['power_watts'] = d.read('uint:8') * 4.0

        goal_lat_bits = BitArray(d.read('bits:24'))
        values['goal_latitude'] = decode_latlon(goal_lat_bits)
        goal_lon_bits = BitArray(d.read('bits:24'))
        values['goal_longitude'] = decode_latlon(goal_lon_bits)

        values['battery_percent'] = d.read('uint:8')

        gfi_pitch_oil_encoded = BitArray(d.read('bits:16'))
        gfi_pitch_oil_encoded.byteswap()
        values['gfi_percent'] = gfi_pitch_oil_encoded[11:].uint * 100.0 / 31.0
        values['oil'] = gfi_pitch_oil_encoded[6:11].uint * 100.0 / 31.0
        values['pitch'] = gfi_pitch_oil_encoded[0:6].int * 180.0 / 63.0

        # Make a message
        mdat_state = cls(values)

        return mdat_state

    @property
    def as_data(self):
        s = bitstring.pack('uint:8, bits:24, bits:24, uint:8, bits:24, uint:8, bits:16, bits:40, uint:8, uint:8, '
                           'uint:8, uint:8, bits:24, bits:24, uint:8, bits:16',
                           14,  # MDAT_STATE
                           encode_latlon(self['latitude']),
                           encode_latlon(self['longitude']),
                           self['fix_age'] // 4,
                           encode_time_data(self['time_date']),
                           self['heading'] * (255.0 / 360.0),
                           encode_mission_mode_and_depth(self['mission_mode'], self['depth']),
                           BitArray(uint=self['faults_bits'], length=40),
                           self['mission_leg'],
                           self['estimated_velocity'] * 25,
                           self['objective_index'],
                           self['power_watts'] // 4,
                           encode_latlon(self['goal_latitude']),
                           encode_latlon(self['goal_longitude']),
                           self['battery_percent'],
                           encode_gfi_oil_pitch(self['gfi_percent'], self['oil'], self['pitch'])
                           )
        return s.bytes


CclTypes = {11: MdatCommandMessage,
            14: MdatStateMessage,
            0x10: MdatRangerMessage}


#!/usr/bin/env python
#__author__ = 'Andrew'

from acomms import micromodem, unifiedlog
import logging
from time import sleep
import argparse

if __name__ == '__main__':
        ap = argparse.ArgumentParser(description ='Connect to a MM for testing purposes')
        ap.add_argument("logpath", help="Location of Log File", default="/home/acomms/")
        ap.add_argument("-C","--COM", help='COM Port to connect', default="/dev/ttyO1")
        ap.add_argument("-BR","--Baudrate", help="COM Port Baud Rate", default=19200)
        

        args = ap.parse_args()

        unified_log = unifiedlog.UnifiedLog(log_path=args.logpath, console_log_level=logging.INFO)

        um1 = micromodem.Micromodem(name='Micromodem2',unified_log=unified_log)

        um1.connect_serial(args.COM, args.Baudrate)

        try:
            while True:
                sleep(1)
    
        finally:
            um1.disconnect()



from acomms import Micromodem, UnifiedLog
from time import sleep

unilog = UnifiedLog(console_log_level='INFO')

modem_a = Micromodem(name="MM1", unified_log=unilog )
modem_a.connect('COM7', 19200)

modem_a.set_config('SRC', 0)

i=0;
send_signals = [41,42,43,44]
receive_signals = [25,26,27,28]
for txsig in send_signals:
    modem_a.send_pgt(txsig,receive_signals[0]+i*5,receive_signals[1]+i*5,receive_signals[2]+i*5,receive_signals[3]+i*5,5000)
    sleep(10)

'''
#REMUS vehicle signals
send_signals = [41,42,43,44]
receive_signals = [21,22,23,24]
i = 0;
for txsig in send_signals:
    modem_a.send_pgt(txsig,receive_signals[0]+i*5,receive_signals[1]+i*5,receive_signals[2]+i*5,receive_signals[3]+i*5,5000)
    i = i+ 1
    sleep(10)

send_signals = [41,42,43,44]
receive_signals = [25,26,27,28]
i = 0;
for txsig in send_signals:
    modem_a.send_pgt(txsig,receive_signals[0]+i*5,receive_signals[1]+i*5,receive_signals[2]+i*5,receive_signals[3]+i*5,5000)
    i = i+ 1
    sleep(10)

send_signals = [40]
receive_signals = [71,72,73,74]
for txsig in send_signals:
    modem_a.send_pgt(txsig,receive_signals[0],receive_signals[1],receive_signals[2],receive_signals[3],5000)
    sleep(10)
    '''


__author__ = 'Eric'

from acomms.modem_connections.modem_connection import ModemConnection
from acomms.modem_connections.iridium_connection import IridiumConnection
from acomms.modem_connections.serial_connection import SerialConnection
from acomms.modem_connections.sbd_connection import SBDEmailConnection
from acomms.modem_connections.tcp_connection import TcpConnection
from acomms.modem_connections.udp_connection import UdpConnection
from acomms.modem_connections.tcp_client_connection import TcpClientConnection





import argparse
from Queue import Queue
from acomms import Micromodem, nmeaChecksum, unifiedlog
from twisted.internet import reactor
from twisted.internet.protocol import ServerFactory
from twisted.protocols.basic import LineReceiver
import logging


class BridgeProtocol(LineReceiver):
    def __init__(self, factory):
        self.factory = factory
        self.modem = self.factory.modem

    def lineReceived(self, line):
        # If we have data here, pass it to the modem.
        self.modem.write_string(line + "\r\n")

    def connectionMade(self):
        self.factory.clients.append(self)
        self.modem._daemon_log.info('TCP client connected ({} total)'.format(len(self.factory.clients)))

    def connectionLost(self, reason):
        self.factory.clients.remove(self)
        self.modem._daemon_log.info('TCP client disconnected ({} total)'.format(len(self.factory.clients)))

class BridgeFactory(ServerFactory):
    protocol = BridgeProtocol

    def buildProtocol(self, address):
        return BridgeProtocol(self)

class ModemBridge(object):
    def __init__(self):
        self.modem = None
        self.factory = None


    def on_modem_message(self, msg):
        # This is a tad ridiculous
        message = ",".join( [str(p) for p in [ msg['type'] ] + msg['params']] )
        chk = nmeaChecksum( message )
        message = "$" + message.lstrip('$').rstrip('\r\n*') + "*" + chk

        for client in self.factory.clients:
            reactor.callFromThread(client.sendLine,message)
        pass

    def connect_to_iridium(self, number, port, baudrate):
        unified_log = unifiedlog.UnifiedLog(console_log_level=logging.INFO)
        self.modem = Micromodem(unified_log=unified_log)
        self.modem.connect_iridium(number, port, baudrate)

        # Now, attach to NMEA message events.
        self.modem.nmea_listeners.append(self.on_modem_message)

    def start_server(self, port):
        self.factory = BridgeFactory()
        # Stick a property on the factory that we can use.
        self.factory.modem = self.modem
        self.factory.clients = []
        #self.factory.protocol = BridgeProtocol

        reactor.listenTCP(port, self.factory)

    def start_bridge(self, number, serial_port, baudrate, tcp_port):
        self.connect_to_iridium(args.iridium_number, args.iridium_serial_port, args.iridium_baud)
        self.start_server(tcp_port)

        reactor.run()




if __name__ == '__main__':
    ap = argparse.ArgumentParser(description='Connect a modem to a socket.')
    #ap.add_argument("-l", "--logfile", help="Save log with specified name")
    ap.add_argument("iridium_number", help="Iridium phone number to dial")
    ap.add_argument("iridium_serial_port", help="Iridium modem serial port")
    ap.add_argument("iridium_baud", help="Iridium modem baud rate")
    ap.add_argument("tcp_server_port", help="Listen on this port")


    args = ap.parse_args()

    bridge = ModemBridge()
    bridge.start_bridge(args.iridium_number, args.iridium_serial_port, int(args.iridium_baud), int(args.tcp_server_port))




from acomms.modem_connections import ModemConnection
from serial import Serial
from time import sleep
from threading import Thread


class SerialConnection(ModemConnection):

    def __init__(self, modem, port, baudrate, timeout=0.1):
        self._incoming_line_buffer = ""

        self.connection_type = "serial"

        self.modem = modem
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout

        self._serialport = Serial(port, baudrate, timeout=self.timeout)

        self._thread = Thread(target=self._listen)
        self._thread.setDaemon(True)
        self._thread.start()

    @property
    def is_connected(self):
        return self._serialport.isOpen()

    @property
    def can_change_baudrate(self):
        return True

    def change_baudrate(self,baudrate):
        self.baudrate = baudrate
        self._serialport.baudrate = baudrate
        return baudrate


    def close(self):
        #self._listen.stop()
        self._serialport.close()

    def _listen(self):
        while True:
            if self._serialport.isOpen():
                msg = self.readline()
                # We are connected, so pass through to NMEA
                self.modem._process_incoming_nmea(msg)
                self.modem._process_outgoing_nmea()
            else: # not connected
                sleep(0.5) # Wait half a second, try again.

    def readline(self):
        """Returns a \n terminated line from the modem.  Only returns complete lines (or None on timeout)"""
        rl = self._serialport.readline().decode('utf-8')

        if rl == "":
            return None

        # Make sure we got a complete line.  Serial.readline may return data on timeout.
        if rl[-1] != '\n':
            self._incoming_line_buffer += rl
            return None
        else:
            if self._incoming_line_buffer != "":
                rl = self._incoming_line_buffer + rl
            self._incoming_line_buffer = ""
            return rl

    def write(self, data):
        self._serialport.write(data.encode('utf-8'))



import binascii
from acomms.cyclestats import CycleStats
from acomms.index_list import IndexList

class FDPacket(object):
    def __init__(self, src, dest, miniframe_rate, dataframe_rate, ack,
                 cyclestats=None, miniframes=None, dataframes=None,
                 minibytes=None, databytes=None, version=7):

        if miniframes is not None and minibytes is not None:
            raise ValueError("FDPacket should only be initialized from received frames or simply the desired bytes")

        if dataframes is not None and databytes is not None:
            raise ValueError("FDPacket should only be initialized from received frames or simply the desired bytes")

        self.src = src
        self.dest = dest
        self.miniframe_rate = miniframe_rate
        self.dataframe_rate = dataframe_rate
        self.ack = ack
        self.version = version

        self.cyclestats = cyclestats

        if miniframes is None and minibytes is None:
            self.miniframes = []

        if dataframes is None and databytes is None:
            self.dataframes = []

        if minibytes is not None:
            self._minibytes = bytes(minibytes)
        else:
            self._minibytes = bytes()

        if databytes is not None:
            self._databytes = bytes(databytes)
        else:
            self._databytes = bytes()

    @property
    def packet_is_good(self):
        if self._minibytes or self._databytes:
            return True
        else:
            packet_is_good = True
            if self.miniframes:
                packet_is_good = packet_is_good and bool(min([miniframe.crccheck for miniframe in self.miniframes]))
            if self.dataframes:
                packet_is_good = packet_is_good and bool(min([dataframe.crccheck for dataframe in self.dataframes]))

            return packet_is_good

    @property
    def minibytes(self):
        if self._minibytes:
            return self._minibytes
        return bytes().join([miniframe.data for miniframe in self.miniframes])

    @property
    def databytes(self):
        if self._databytes:
            return self._databytes
        return bytes().join([frame.data for frame in self.dataframes])

    @property
    def valid_minibytes(self) -> IndexList:
        return self.get_valid_index(self.miniframes)

    @property
    def valid_databytes(self) -> IndexList:
        return self.get_valid_index(self.dataframes)

    def get_valid_index(self, frames) -> IndexList:
        valid_index = IndexList()

        idx = 0

        if len(frames) > 0:
            for frame in frames:
                length = frame.nbytes
                start = idx
                end = idx + length

                if frame.crccheck:
                    valid_index.append([start, end])

                idx = idx + length

            #valid_index.simplify()

        return valid_index

    def parse_frames(self, framestring):
        framelist = []

        framesplit = framestring.split(';')
        nframes = int(len(framesplit)/3)

        for i in range(0, nframes):
            framelist.append(FDFrame(int(framesplit[i*3]) == 1,
                                     int(framesplit[i*3+1]),
                                     bytearray.fromhex(framesplit[i*3+2])))

        return framelist

    def load_legacy_frames(self, legacyframes):
        framelist = []

        nframes = int(len(legacyframes))

        for i in range(0, nframes):
            framelist.append(FDFrame(~legacyframes[i].bad_crc, len(legacyframes[i].data), legacyframes[i].data))

        self.dataframes = framelist

    def add_miniframes(self, miniframestring):
        self.miniframes = self.miniframes + self.parse_frames(miniframestring)

    def add_dataframes(self, dataframestring):
        self.dataframes = self.dataframes + self.parse_frames(dataframestring)

    def add_cyclestats(self, cyclestats):
        self.cyclestats = cyclestats

    def msg_tfp(self, ver_pass):
        if ver_pass:
            msg = {'type': 'CCTFP', 'params': [0,
                                               0,
                                               self.dest,
                                               self.miniframe_rate,
                                               self.dataframe_rate,
                                               self.ack,
                                               0,
                                               self._minibytes.hex(),
                                               self._databytes.hex()]}

        else:
            msg = {'type': 'CCTFP', 'params': [self.dest,
                                               self.miniframe_rate,
                                               self.dataframe_rate,
                                               self.ack,
                                               0,
                                               self._minibytes.hex(),
                                               self._databytes.hex()]}

        return msg

    def msg_tdp(self):
        msg = {'type': 'CCTDP', 'params': [self.dest,
                                           self.dataframe_rate,
                                           0,
                                           0,
                                           self._databytes.hex()]}
        return msg

class FDFrame(object):
    def __init__(self, crccheck, nbytes, data):
        self.crccheck = crccheck
        self.nbytes = nbytes
        self.data = bytearray(data)



__author__ = 'Eric'

from datetime import datetime, timedelta
import isodate


def convert_to_datetime(object_to_convert):
    # First, is it already a datetime?
    if isinstance(object_to_convert, datetime):
        as_datetime = object_to_convert
    else:
        # If it is an decimal (or sting convertable to a decimal), assume it is a unix time.
        try:
            as_datetime = datetime.utcfromtimestamp(int(object_to_convert))
        except ValueError:
            # If it can't be an int, try to parse it as a string.  Blank strings return None
            try:
                if object_to_convert == "":
                    return None
                as_datetime = isodate.parse_datetime(object_to_convert)
            except ValueError:
                raise(ValueError("Unable to parse object into a datetime"))

    return as_datetime

def convert_to_timedelta(object_to_convert):
    # Is it already a timeval?
    if isinstance(object_to_convert, timedelta):
        as_timedelta = object_to_convert
    else:
        # If it is an number (or a number as string or something), assume it is in seconds.
        try:
            as_timedelta = timedelta(seconds=int(object_to_convert))
        except ValueError:
            # Try parsing it as an ISO8601 string
            try:
                as_timedelta = isodate.parse_duration(object_to_convert)
            except ValueError:
                raise(ValueError("Unable to parse object into a timedelta"))
    return as_timedelta


def to_utc_iso8601(datetime_to_convert, strip_fractional_seconds=False):
    # Strip fractional seconds, if requested.
    if strip_fractional_seconds:
        datetime_to_convert.replace(microsecond=0)

    if (datetime_to_convert.tzinfo is None) or (datetime_to_convert.utcoffset is None):
        datetime_to_convert = datetime_to_convert.replace(tzinfo=isodate.UTC)
    else:
        datetime_to_convert = datetime_to_convert.astimezone(isodate.UTC)

    return isodate.datetime_isoformat(datetime_to_convert)





from collections import UserList

class IndexList(UserList):
    ''' Class to contain list of start/end index pairs'''
    def __init__(self, index_list=None):
        if index_list is not None:
            if not any(isinstance(i, list) for i in index_list):
                if len(index_list) == 2:
                    self.data = [index_list]
                elif len(index_list) == 0:
                    self.data = []
                else:
                    raise Exception('index_list argument must be one or more [start,end] index pairs')
            elif all(isinstance(i, list) for i in index_list) and all((len(i) == 2) for i in index_list):
                self.data = index_list
            else:
                raise Exception('index_list argument must be one or more [start,end] index pairs')
        else:
            self.data = []

    @classmethod
    def from_list(cls, flatlist):
        data = []

        # Opposite of flat, takes a flat list and turns it into 2d index list
        if len(flatlist) % 2:
            raise Exception('list argument must have even number of entry pairs')

        for i in range(0, len(flatlist), 2):
            if flatlist[i] > flatlist[i+1]:
                raise Exception('list argument must be in ordered start:end pairs')

            data.append([flatlist[i], flatlist[i+1]])

        return cls(data)

    @property
    def flat(self):
        flat = []
        for pair in self.data:
            flat.append(pair[0])
            flat.append(pair[1])

        if len(flat) % 2:
            raise Exception('IndexList has odd number of elements')

        return flat

    def append(self, items):
        #If list is a list of multiple start/end pairs
        if all(isinstance(i, list) for i in items):
            for i in items:
                if len(i) != 2:
                    raise Exception('list argument must be one or more [start,end] index pairs')
                self._safe_append(i)
        else:
            if len(items) != 2:
                raise Exception('list argument must be one or more [start,end] index pairs')
            self._safe_append(items)

    def _safe_append(self, new_index):
        self.simplify()

        expanded = self._expand()
        expanded = expanded + list(range(new_index[0], new_index[-1]))
        self.data = self._contract(expanded)

    def inverse(self, limits):
        if any(isinstance(i, list) for i in limits):
            raise Exception('Limits cannot be nested list')
        if len(limits) != 2:
            raise Exception('Limits must be single [start,end] pair')

        inverse_index = []

        self.sort()

        #Check empty
        if not self.data:
            inverse_index = limits
        else:
            # Check beginning
            if self.data[0][0] > limits[0]:
                inverse_index.append([limits[0], self.data[0][0]])

            # Check middle
            if len(self.data) > 1:
                for i in range(len(self.data) - 2):
                    inverse_index.append([self.data[i][-1], self.data[i+1][0]])

            # Check end
            if self.data[-1][-1] < limits[-1]:
                inverse_index.append([self.data[-1][-1], limits[-1]])

        self._simplify(inverse_index)

        return inverse_index

    def simplify(self):
        if len(self.data) > 0:
            self.data = self._simplify(self.data)

        return self.data

    def _simplify(self, index_list):
        i = 0

        index_list.sort()

        while i < len(index_list) - 1:
            a = index_list[i][0]
            b = index_list[i][-1]
            c = index_list[i + 1][0]
            d = index_list[i + 1][-1]
            if b == c:
                index_list[i:i + 2] = [[a, d]]
            else:
                i = i + 1

        return index_list

    def _expand(self):
        expanded = []

        for index in self.data:
            expanded = expanded + list(range(index[0], index[1]))

        return expanded

    def _ranges(self, nums):
        nums = sorted(set(nums))
        gaps = [[s+1, e] for s, e in zip(nums, nums[1:]) if s + 1 < e]
        begin = nums[:1]
        end = [nums[-1:][0]+1]
        edges = iter(begin + sum(gaps, []) + end)

        return list(map(list, zip(edges, edges)))

    def _contract(self, expanded_form):
        expanded_form.sort()

        expanded_form = list(dict.fromkeys(expanded_form))

        return self._ranges(expanded_form)





import os
import sys
from time import sleep, time
from datetime import datetime
from acomms import timeutil
import re
from queue import Queue, Empty, Full
import logging
import struct
from collections import namedtuple
import hashlib
from binascii import hexlify
from serial import Serial
import binascii
import ctypes
import functools

import acomms.commstate
from acomms.messageparser import MessageParser
from acomms.messageparams import Packet, CycleInfo, hexstring_from_data, Rates, DataFrame, FDPMiniRates, FDPDataRates,LDRRates
from acomms.modem_connections import SerialConnection
from acomms.modem_connections import IridiumConnection
from acomms.modem_connections import SBDEmailConnection
from acomms.flexibledataprotocol import FDPacket
from acomms.unifiedlog import UnifiedLog
try:
    from acomms.REMUSnav import dtcodes
except ImportError:
    pass

# Convert a string to a byte listing
toBytes = lambda inpStr: map(ord, inpStr)
# Convert a list to a hex string (each byte == 2 hex chars)
toHexStr = lambda inpLst: "".join(["%02X" % x for x in inpLst])
# Calculate hex-encoded, XOR checksum for input string per NMEA 0183.
nmeaChecksum = lambda inpStr: toHexStr([functools.reduce(lambda x, y: x ^ y, toBytes(inpStr))])
# Convert boolean to C / CCL / Modem representation (0,1)
bool2int = lambda inbool: inbool and 1 or 0


def nmeaChecksum32(inpStr):
    crc32tabr = [
        0x00000000, 0x8c8cd047, 0xb5dec035, 0x39521072, 0xc77ae0d1, 0x4bf63096, 0x72a420e4, 0xfe28f0a3,
        0x2232a119, 0xaebe715e, 0x97ec612c, 0x1b60b16b, 0xe54841c8, 0x69c4918f, 0x509681fd, 0xdc1a51ba,
        0x44654232, 0xc8e99275, 0xf1bb8207, 0x7d375240, 0x831fa2e3, 0x0f9372a4, 0x36c162d6, 0xba4db291,
        0x6657e32b, 0xeadb336c, 0xd389231e, 0x5f05f359, 0xa12d03fa, 0x2da1d3bd, 0x14f3c3cf, 0x987f1388,
        0x88ca8464, 0x04465423, 0x3d144451, 0xb1989416, 0x4fb064b5, 0xc33cb4f2, 0xfa6ea480, 0x76e274c7,
        0xaaf8257d, 0x2674f53a, 0x1f26e548, 0x93aa350f, 0x6d82c5ac, 0xe10e15eb, 0xd85c0599, 0x54d0d5de,
        0xccafc656, 0x40231611, 0x79710663, 0xf5fdd624, 0x0bd52687, 0x8759f6c0, 0xbe0be6b2, 0x328736f5,
        0xee9d674f, 0x6211b708, 0x5b43a77a, 0xd7cf773d, 0x29e7879e, 0xa56b57d9, 0x9c3947ab, 0x10b597ec,
        0xbd526873, 0x31deb834, 0x088ca846, 0x84007801, 0x7a2888a2, 0xf6a458e5, 0xcff64897, 0x437a98d0,
        0x9f60c96a, 0x13ec192d, 0x2abe095f, 0xa632d918, 0x581a29bb, 0xd496f9fc, 0xedc4e98e, 0x614839c9,
        0xf9372a41, 0x75bbfa06, 0x4ce9ea74, 0xc0653a33, 0x3e4dca90, 0xb2c11ad7, 0x8b930aa5, 0x071fdae2,
        0xdb058b58, 0x57895b1f, 0x6edb4b6d, 0xe2579b2a, 0x1c7f6b89, 0x90f3bbce, 0xa9a1abbc, 0x252d7bfb,
        0x3598ec17, 0xb9143c50, 0x80462c22, 0x0ccafc65, 0xf2e20cc6, 0x7e6edc81, 0x473cccf3, 0xcbb01cb4,
        0x17aa4d0e, 0x9b269d49, 0xa2748d3b, 0x2ef85d7c, 0xd0d0addf, 0x5c5c7d98, 0x650e6dea, 0xe982bdad,
        0x71fdae25, 0xfd717e62, 0xc4236e10, 0x48afbe57, 0xb6874ef4, 0x3a0b9eb3, 0x03598ec1, 0x8fd55e86,
        0x53cf0f3c, 0xdf43df7b, 0xe611cf09, 0x6a9d1f4e, 0x94b5efed, 0x18393faa, 0x216b2fd8, 0xade7ff9f,
        0xd663b05d, 0x5aef601a, 0x63bd7068, 0xef31a02f, 0x1119508c, 0x9d9580cb, 0xa4c790b9, 0x284b40fe,
        0xf4511144, 0x78ddc103, 0x418fd171, 0xcd030136, 0x332bf195, 0xbfa721d2, 0x86f531a0, 0x0a79e1e7,
        0x9206f26f, 0x1e8a2228, 0x27d8325a, 0xab54e21d, 0x557c12be, 0xd9f0c2f9, 0xe0a2d28b, 0x6c2e02cc,
        0xb0345376, 0x3cb88331, 0x05ea9343, 0x89664304, 0x774eb3a7, 0xfbc263e0, 0xc2907392, 0x4e1ca3d5,
        0x5ea93439, 0xd225e47e, 0xeb77f40c, 0x67fb244b, 0x99d3d4e8, 0x155f04af, 0x2c0d14dd, 0xa081c49a,
        0x7c9b9520, 0xf0174567, 0xc9455515, 0x45c98552, 0xbbe175f1, 0x376da5b6, 0x0e3fb5c4, 0x82b36583,
        0x1acc760b, 0x9640a64c, 0xaf12b63e, 0x239e6679, 0xddb696da, 0x513a469d, 0x686856ef, 0xe4e486a8,
        0x38fed712, 0xb4720755, 0x8d201727, 0x01acc760, 0xff8437c3, 0x7308e784, 0x4a5af7f6, 0xc6d627b1,
        0x6b31d82e, 0xe7bd0869, 0xdeef181b, 0x5263c85c, 0xac4b38ff, 0x20c7e8b8, 0x1995f8ca, 0x9519288d,
        0x49037937, 0xc58fa970, 0xfcddb902, 0x70516945, 0x8e7999e6, 0x02f549a1, 0x3ba759d3, 0xb72b8994,
        0x2f549a1c, 0xa3d84a5b, 0x9a8a5a29, 0x16068a6e, 0xe82e7acd, 0x64a2aa8a, 0x5df0baf8, 0xd17c6abf,
        0x0d663b05, 0x81eaeb42, 0xb8b8fb30, 0x34342b77, 0xca1cdbd4, 0x46900b93, 0x7fc21be1, 0xf34ecba6,
        0xe3fb5c4a, 0x6f778c0d, 0x56259c7f, 0xdaa94c38, 0x2481bc9b, 0xa80d6cdc, 0x915f7cae, 0x1dd3ace9,
        0xc1c9fd53, 0x4d452d14, 0x74173d66, 0xf89bed21, 0x06b31d82, 0x8a3fcdc5, 0xb36dddb7, 0x3fe10df0,
        0xa79e1e78, 0x2b12ce3f, 0x1240de4d, 0x9ecc0e0a, 0x60e4fea9, 0xec682eee, 0xd53a3e9c, 0x59b6eedb,
        0x85acbf61, 0x09206f26, 0x30727f54, 0xbcfeaf13, 0x42d65fb0, 0xce5a8ff7, 0xf7089f85, 0x7b844fc2]
    crc = 0xffffffff
    bytestring = bytearray(inpStr.encode("iso-8859-1"))
    for i in range(len(bytestring)):
        crc = (crc >> 8) ^ crc32tabr[0xff & (crc ^ bytestring[i])]
    return hex(ctypes.c_uint32(~crc).value)[2:-1].upper()


class ChecksumException(Exception):
    pass


class UnavailableInApiLevelError(Exception):
    pass


class ModemResponseTimeout(Exception):
    pass


class ModemResponseError(Exception):
    pass


class Micromodem(object):
    def __init__(self, name='modem', unified_log=None, log_path=None, log_level='INFO'):

        name = str(name)
        # Strip non-alphanumeric characters from name
        self.name = ''.join([ch for ch in name if ch.isalnum()])

        self.connection = None

        self.nmea_listeners = []

        self.parser = MessageParser(self)
        self.state = acomms.commstate.Idle(modem=self)

        self.rxframe_listeners = []
        self.packetrx_listeners = []
        self.fdp_listeners = []
        self.cst_listeners = []
        self.xst_listeners = []
        self.ack_listeners = []
        self.ping_listeners = []
        self.ping_reply_listeners = []
        self.mpr_listeners = [] #TODO: implement listener
        self.camua_listeners = []
        self.causb2_listeners = []
        self.casst_listeners = []
        self.carev_listeners = []

        self.incoming_dataframe_queues = []
        self.packetrx_queues = []
        self.incoming_all_packet_queues = []
        self.incoming_fdp_queues = []
        self.incoming_cst_queues = []
        self.incoming_xst_queues = []
        self.incoming_msg_queues = []
        self.incoming_mpr_queues = [] #TODO: implement queue
        self.incoming_causb2_queues = []
        self.incoming_casst_queues = []
        self.incoming_ping_queues = []
        self.incoming_ping_reply_queues = []

        self._api_level = 11
        self.default_nmea_timeout = 1

        self.fw_version = None

        self.id = -1
        self.asd = False
        self.pcm_on = False

        # state tracking variables
        self.current_rx_frame_num = 1
        self.current_tx_frame_num = 1
        self.current_cycleinfo = None
        self.current_txpacket = None
        self.current_rxpacket = None
        self.current_fdpacket = None
        self.current_legacy_packet = None
        self.set_host_clock_flag = False

        self.serial_tx_queue = Queue()

        self.get_uplink_data_function = None

        # Use this to work around 7-bit/8-bit addressing issues
        self._ignore_cacyc_address_mismatch = False

        # Set up logging
        #NOTE -- THIS is where the vast majority of the messages in the test script logs are generated
        #So in order to write some kind of code that will tie a $CACST log message from here to the generation of
        #A database entry in Django via the ResultsCycleStats model, it's gonna have to (at least in part) go here
        if unified_log is None:
            unified_log = UnifiedLog(log_path=log_path)
        self._daemon_log = unified_log.getLogger("daemon.{0}".format(self.name))
        self._daemon_log.setLevel(log_level)
        self._nmea_in_log = unified_log.getLogger("nmea.from.{0}".format(self.name))
        self._nmea_in_log.setLevel(logging.INFO)
        self._nmea_out_log = unified_log.getLogger("nmea.to.{0}".format(self.name))
        self._nmea_out_log.setLevel(logging.INFO)
        self.unified_log = unified_log
        self.config_data = {}

    @property
    def api_level(self):
        return self._api_level

    def force_api_level(self, api_level):
        self._api_level = api_level

    def version_a_at_least_b(self, verstr_a, verstr_b):
        verstr_regex = re.compile(
            "v?(?P<verstr>(?P<major>\d+).(?P<minor>\d+).(?P<patch>\d+))(-rc(?P<rc>\d+))?(-(?P<ncommits>\d+)-g(?P<hash>\w*))?")

        a = verstr_regex.search(verstr_a)
        b = verstr_regex.search(verstr_b)

        for idx in [2, 3, 4]:
            a_val = int(a.group(idx))
            b_val = int(b.group(idx))
            if a_val == b_val:
                continue
            else:
                return a_val >= b_val

        a_rc = int((a.group('rc') or 999999))
        b_rc = int((b.group('rc') or 999999))

        if a_rc == b_rc:
            a_ncommits = int((a.group('ncommits') or 0))
            b_ncommits = int((b.group('ncommits') or 0))
            return a_ncommits >= b_ncommits
        else:
            return a_rc >= b_rc

    def fw_at_least(self, reference_ver):
        if not self.fw_version:
            self.write_nmea("$CCCFQ,info")
            sleep(1)
            if not self.fw_version:
                self.fw_version = 'v0.0.0'

        return self.version_a_at_least_b(self.fw_version, reference_ver)

    def connect(self, serialport=None, baudrate=19200):
        ''' Convenience function to establish a connection using a serial port.  Included for backward-compatibility.
        '''
        self.connect_serial(serialport, baudrate)

    def connect_serial(self, port, baudrate=19200):
        self.connection = SerialConnection(self, port, baudrate)
        self._daemon_log.info("Connected to {0} ({1} bps)".format(port, baudrate))
        sleep(0.05)
        self.get_config('SRC')
        # self.query_modem_info()
        # self.query_nmea_api_level()

    def connect_iridium(self, number, port, baudrate=19200):
        self.connection = IridiumConnection(modem=self, port=port, baudrate=baudrate, number=number)
        self._daemon_log.info("Connected to Iridium #:{0} on Serial {1}({2} bps)".format(number, port, baudrate))

    def connect_sbd_email(self, IMEI, email_account='acomms-sbd@whoi.edu',
                          username=None, pw=None,
                          check_rate_sec=30,
                          imap_srv="imap.whoi.edu", imap_port=143,
                          smtp_svr="outbox.whoi.edu", smtp_port=25, DoD=False):
        self.connection = SBDEmailConnection(modem=self,
                                             IMEI=IMEI,
                                             email_account=email_account,
                                             username=username, pw=pw,
                                             check_rate_min=check_rate_sec / 60,
                                             imap_srv=imap_srv, imap_port=imap_port,
                                             smtp_svr=smtp_svr, smtp_port=smtp_port,
                                             DoD=DoD)
        self._daemon_log.info(
            "Using Email Based SBD Connection (IMEI#: {0} Address: {1} IMAP Server: {2}:{3} SMTP Server: {4}:{5} Checking Every {6} Minutes)".format(
                IMEI, email_account, imap_srv, imap_port, smtp_svr, smtp_port, check_rate_sec / 60))

    def disconnect(self):
        if self.connection is not None:
            self.connection.close()
            self.connection = None

    def query_modem_info(self):
        sleep(0.05)
        self.get_config_param("SRC")
        sleep(0.05)
        # All of the salient properties on this object are populated automatically by the NMEA config handler.

    def query_nmea_api_level(self):
        api_level = 0

        msg = {'type': "CCALQ", 'params': [0]}
        self.write_nmea(msg)

        response = self.wait_for_nmea_type('CAALQ', timeout=5)

        if response:
            if response['type'] == 'CAALQ':
                api_level = int(response['params'][1])
            else:
                api_level = 0
        else:
            api_level = 0

        self._api_level = api_level
        return api_level

    def set_nmea_api_level(self, api_level=11):

        self._api_level = api_level
        return api_level

    def start_nmea_logger(self, nmea_log_format, log_path=None, file_name=None):
        """ This starts an additional log file that only logs NMEA messages, primarily for compatibility with other
            systems.
        """
        nmea_log_format = nmea_log_format.lower()
        if nmea_log_format == "modemtma":
            logformat = logging.Formatter("%(asctime)sZ,RX,%(message)s", "%Y-%m-%d %H:%M:%S")
        elif nmea_log_format == "messagesonly":
            logformat = logging.Formatter("%(message)s")
        elif nmea_log_format == "timestamped":
            logformat = logging.Formatter("%(asctime)s\t%(message)s", "%Y-%m-%dT%H:%M:%SZ")
        else:
            raise (ValueError("Unrecognized nmea_log_format"))

        # If no log path is specified, use (or create) a directory in the user's home directory
        if log_path is None:
            log_path = os.path.expanduser('~/acomms_logs')
        log_path = os.path.normpath(log_path)

        # Create the directory if it doesn't exist
        if not os.path.isdir(log_path):
            os.makedirs(log_path)

        if file_name is None:
            now = datetime.utcnow()
            file_name = "nmea_{0}_{1}.log".format(self.name, now.strftime("%Y%m%dT%H%M%SZ"))

        log_file_path = os.path.join(log_path, file_name)

        handler = logging.FileHandler(log_file_path)
        handler.setLevel(logging.INFO)

        self._nmea_in_log.addHandler(handler)

        self._daemon_log.debug("Started new NMEA log")

    def _process_outgoing_nmea(self):
        # Now, transmit anything we have in the outgoing queue.
        try:
            txstring = self.serial_tx_queue.get_nowait()
            self.connection.write(txstring)
            self._nmea_out_log.info(txstring.rstrip('\r\n'))
        # If the queue is empty, then pass, otherwise log error
        except Empty:
            pass
        except:
            self._daemon_log.exception("NMEA Output Error")

    def _process_incoming_nmea(self, msg):
        if msg is not None:
            try:
                # self.nmealog.info("< " + msg.rstrip('\r\n'))
                self._nmea_in_log.info(msg.rstrip('\r\n'))

                msg = Message(msg)

                self.parser.parse(msg)

                try:
                    for func in self.nmea_listeners:
                        func(msg)  # Pass the message to any custom listeners.
                except Exception as e:
                    self._daemon_log.warn("Error in custom listener: ")
                    self._daemon_log.warn(repr(e))

                # Append this message to all listening queues
                for q in self.incoming_msg_queues:
                    try:
                        q.put_nowait(msg)
                    except:
                        self._daemon_log.warn("Error appending to incoming message queue")
            except ChecksumException:
                self._daemon_log.warn("NMEA Checksum Error: %s" % (msg.rstrip('\r\n')))
            except:
                self._daemon_log.warn("NMEA Input Error")

    def _message_waiting(self):
        return not self.serial_tx_queue.empty()

    def close_loggers(self):
        for hdlr in self._daemon_log.handlers:
            hdlr.flush()
            hdlr.close()
            self._daemon_log.removeHandler(hdlr)

        for hdlr in self._nmea_in_log.handlers:
            hdlr.flush()
            hdlr.close()
            self._nmea_in_log.removeHandler(hdlr)

        for hdlr in self._nmea_out_log.handlers:
            hdlr.flush()
            hdlr.close()
            self._nmea_out_log.removeHandler(hdlr)

    def _changestate(self, newstate):
        self.state = newstate(modem=self)
        self._daemon_log.debug("Changed state to " + str(self.state))
        self.state.entering()

    def write_nmea(self, msg):
        """Call with the message to send, as an NMEA message.  Correct checksum will be computed."""

        if type(msg) == str:
            # Automagically convert it into an NMEA message (or try, at least)
            msg = Message(msg)

        message = ",".join([str(p) for p in [msg['type']] + msg['params']])
        chk = nmeaChecksum(message)
        message = "$" + message.lstrip('$').rstrip('\r\n*') + "*" + chk + "\r\n"
        # print message # for debug.
        # Serial.write(self, message )

        # Queue this message for transmit in the serial thread
        self._daemon_log.debug("Writing NMEA to output queue: %s" % (message.rstrip('\r\n')))
        try:
            self.serial_tx_queue.put(message, block=False)
        # If queue full, then ignore
        except Full:
            self._daemon_log.debug("write_nmea: Serial TX Queue Full")

    def write_string(self, string):
        self._daemon_log.debug("Writing string to output queue: %s" % (string.rstrip('\r\n')))
        try:
            self.serial_tx_queue.put(string, block=False)
        # If queue full, then ignore
        except Full:
            self._daemon_log.debug("write_string: Serial TX Queue Full")

    def get_config_param(self, param):
        msg = {'type': "CCCFQ", 'params': [param]}
        self.write_nmea(msg)

    def get_config(self, param, response_timeout=2):
        msg = {'type': "CCCFQ", 'params': [param]}
        self.write_nmea(msg)

        config_dict = {}

        if response_timeout:
            while True:
                msg = self.wait_for_nmea_type('CACFG', timeout=response_timeout)
                if msg:
                    param_name = msg['params'][0]
                    param_value = msg['params'][1]
                    self.config_data[param_name] = param_value
                    # If we queried a config group, there is no associated value to add to the dictionary.
                    if param_value != "":
                        config_dict[param_name] = param_value
                    # We aren't done unless this config parameter matched the one we queried
                    if param_name == param:
                        break
                    else:
                        continue
                else:
                    break

            if config_dict:
                return config_dict
            else:
                return None
        else:
            return None

    def raw_readline(self):
        """Returns a raw message from the modem."""
        rl = Serial.readline(self)

        if rl == "":
            return None

        # Make sure we got a complete line.  Readline will return data on timeout.
        if rl[-1] != '\n':
            self.temp_incoming_nmea += rl
            return None
        else:
            if self.temp_incoming_nmea != "":
                rl = self.temp_incoming_nmea + rl
            self.temp_incoming_nmea = ""

        return rl

    def on_rxframe(self, dataframe):
        self._daemon_log.debug("I got a frame!  Yay!")
        for func in self.rxframe_listeners:
            func(dataframe)
        # Append this message to all listening queues
        for q in self.incoming_dataframe_queues:
            try:
                q.put_nowait(dataframe)
            except:
                self._daemon_log.warn("Error appending to incoming dataframe queue")

    def on_minipacket_tx_failed(self):
        self._daemon_log.warn("Minipacket transmit failed.")

    def on_packettx_failed(self):
        self._daemon_log.warn("Packet transmit failed.")

    def on_packettx_success(self):
        self._daemon_log.info("Packet transmitted successfully")

    def on_packetrx_failed(self):
        self._daemon_log.warn("Packet RX failed")

    def on_packetrx_success(self, current_packet):
        self._daemon_log.info("Packet RX succeeded")
        # Callbacks
        self.current_legacy_packet = FDPacket(current_packet.cycleinfo.src, current_packet.cycleinfo.dest, None, current_packet.cycleinfo.rate_num, current_packet.cycleinfo.ack)
        self.current_legacy_packet.load_legacy_frames(current_packet.frames)

        for func in self.packetrx_listeners:
            func(current_packet)
        # Append this message to all listening queues
        for q in self.packetrx_queues:
            try:
                q.put_nowait(current_packet)
            except:
                self._daemon_log.warn("Error appending to incoming dataframe queue")

    def on_ack(self, ack, msg):
        self._daemon_log.debug("Got ACK message")
        for func in self.ack_listeners:
            func(ack, msg)  # Pass on the ACK message.

    def on_cst(self, cst, msg):
        self._daemon_log.debug("Got CST message")

        for func in self.cst_listeners:
            func(cst, msg)  # Pass on the CST message.

        # Append this message to all listening queues
        for q in self.incoming_cst_queues:
            try:
                q.put_nowait(cst)
            except:
                self._daemon_log.warn("Error appending to incoming CST queue")

        # Attach CST to an in-progress FDP reception
        # Hacked way to complete an FDP reception with
        # cycle stats included
        if self.current_fdpacket is not None:
            self.current_fdpacket.add_cyclestats(cst)

            # Raise FDP receive event
            self.on_fdp(self.current_fdpacket)

            # Reset FDP state
            self.current_fdpacket = None

        # Attach CST to an in-progress legacy reception
        # Hacked way to complete an legacy reception with
        # cycle stats included
        if self.current_legacy_packet is not None:
            self.current_legacy_packet.add_cyclestats(cst)

            # Raise FDP receive event
            self.on_legacy_rx(self.current_legacy_packet)

            # Reset FDP state
            self.current_legacy_packet = None

    def on_causb2(self, causb2, msg):
        self._daemon_log.debug("Got CAUSB2 message")

        for func in self.causb2_listeners:
            func(causb2, msg)

        for q in self.incoming_causb2_queues:
            try:
                q.put_nowait(causb2)
            except:
                self._daemon_log.warn("Error appending to incoming CAUSB2 queue")

    def on_casst(self, casst, msg):
        self._daemon_log.debug("Got CASST message")

        for func in self.casst_listeners:
            func(casst, msg)

        for q in self.incoming_casst_queues:
            try:
                q.put_nowait(casst)
            except:
                self._daemon_log.warn("Error appending to incoming CASST queue")

    def on_carev(self, msg):
        for func in self.carev_listeners:
            func(msg)

    def on_ping(self, ping, msg):
        self._daemon_log.debug("Got PING")
        if ping.is_reply():
            for q in self.incoming_ping_reply_queues:
                try:
                    q.put_nowait(ping)
                except Exception as ex:
                    self._daemon_log.error("Error appending to incoming ping queue: " + str(sys.exc_info()[0]))

            for func in self.ping_reply_listeners:
                func(ping, msg)
        else:
            for q in self.incoming_ping_queues:
                try:
                    q.put_nowait(ping)
                except Exception as ex:
                    self._daemon_log.error("Error appending to incoming ping queue: " + str(sys.exc_info()[0]))

            for func in self.ping_listeners:
                func(ping, msg)

    def on_legacy_rx(self,legacy_packet):
        self._daemon_log.info("Got Legacy packet")

        # Append packet to all listening queues
        for q in self.incoming_all_packet_queues:
            try:
                q.put_nowait(legacy_packet)
            except:
                self._daemon_log.warn("Error appending to incoming legacy queue")

    def on_fdp(self, fdpacket):
        self._daemon_log.info("Got FDP packet")

        for func in self.fdp_listeners:
            func(fdpacket)

        # Append packet to all listening queues
        for q in self.incoming_fdp_queues:
            try:
                q.put_nowait(fdpacket)
            except:
                self._daemon_log.warn("Error appending to incoming FDP queue")

        for q in self.incoming_all_packet_queues:
            try:
                q.put_nowait(fdpacket)
            except:
                self._daemon_log.warn("Error appending to incoming all packet queue")


    def on_xst(self, xst, msg):
        self._daemon_log.debug("Got XST message")

        for func in self.xst_listeners:
            func(xst, msg)  # Pass on the CST message.

        # Append this message to all listening queues
        for q in self.incoming_xst_queues:
            try:
                q.put_nowait(xst)
            except:
                self._daemon_log.warn("Error appending to incoming XST queue")

    def on_campr(self, campr):
        self._daemon_log.debug("CAMPR received.")
        for func in self.mpr_listeners:
            func(campr)

    def on_camua(self, camua):
        self._daemon_log.debug("Minipacket received")
        for func in self.camua_listeners:
            func(camua)

    def on_log_msg(self, log, msg):
        self._daemon_log.debug("Got Logged message")

        for func in self.log_listeners:
            func(log, msg)  # Pass on the Log message.

        # Append this message to all listening queues
        for q in self.incoming_log_queues:
            try:
                q.put_nowait(log)
            except:
                self._daemon_log.warn("Error appending to incoming modem log queue")

    def send_fdpacket(self, fdpacket, use_tdp=False):
        self.state.send_fdpacket(fdpacket, use_tdp)

    def send_packet(self, packet):
        # FIXME this is a hack
        self.state.send_packet(packet)

    def send_packet_frames(self, dest, rate_num, frames):
        cycleinfo = CycleInfo(self.id, dest, rate_num, False, len(frames))
        packet = Packet(cycleinfo, frames)

        self.send_packet(packet)

    def send_packet_data(self, dest, databytes, rate_num=1, ack=False):
        # When life gives you data, make frames.
        rate = Rates[rate_num]
        src = self.id

        # For now, truncate the data to fit in this packet
        databytes = databytes[0:(rate.maxpacketsize)]

        # Now, make frames.
        frames = []
        for framenum in range(rate.numframes):
            startidx = (framenum * rate.framesize)
            endidx = startidx + rate.framesize
            thisdata = databytes[startidx:endidx]

            if len(thisdata) > 0:
                thisframe = DataFrame(src, dest, ack, framenum, thisdata)
                frames.append(thisframe)

        self.send_packet_frames(dest, rate_num, frames)

    def send_test_packet(self, dest, rate_num=1, num_frames=None, ack=False): #**LOOK HERE FOR ADDING TESTS
        rate = Rates[rate_num]
        src = self.id

        # How many frames shall we send?
        if num_frames == None:
            num_frames = rate.numframes

        # Make frames
        frames = []
        for framenum in range(num_frames):
            framedata = bytearray(struct.pack('!BBBBi', 0, 0, 1, 0, int(time())))
            frame = DataFrame(src, dest, ack, framenum, framedata)
            frames.append(frame)

        # Send a packet
        self.send_packet_frames(dest, rate_num, frames)
        return frames

    def _send_current_txframe(self, frame_num=None):
        # TODO: Make sure we actually have a packet to send

        if frame_num == None:
            frame_num = self.current_tx_frame_num

        self.send_frame(self.current_txpacket.frames[frame_num - 1])

    def send_frame(self, dataframe):
        # Build the corresponding CCTXD message
        msg = {'type': 'CCTXD', 'params': [dataframe.src, dataframe.dest, int(dataframe.ack),
                                           hexstring_from_data(dataframe.data)]}
        self.write_nmea(msg)

    def send_cycleinit(self, cycleinfo):
        self._daemon_log.debug(
            "Sending CCCYC with Following Parameters: %s" % (str([0, cycleinfo.src, cycleinfo.dest, cycleinfo.rate_num,
                                                                  int(cycleinfo.ack), cycleinfo.num_frames])))
        # Build the corresponding CCCYC message
        msg = {'type': 'CCCYC', 'params': [0, cycleinfo.src, cycleinfo.dest, cycleinfo.rate_num,
                                           int(cycleinfo.ack), cycleinfo.num_frames]}

        self.write_nmea(msg)

    def send_uplink_request(self, src_id, dest_id=None, rate_num=1, ack=False):
        if dest_id is None:
            dest_id = self.id

        # The number of frames isn't transmitted acoustically, so it doesn't matter
        cycleinfo = CycleInfo(src_id, dest_id, rate_num, ack, 1)

        self.send_cycleinit(cycleinfo)

    def send_uplink_frame(self, drqparams):
        if self.get_uplink_data_function is not None:
            data = self.get_uplink_data_function(drqparams)
        else:
            data = bytearray(struct.pack('!BBBBi', 0, 0, 1, 0, int(time())))

        self.send_frame(
            DataFrame(src=drqparams.src, dest=drqparams.dest, ack=drqparams.ack, frame_num=drqparams.frame_num,
                      data=data))

    def send_ping(self, dest_id):
        # Build the CCMPC message
        msg = {'type': 'CCMPC', 'params': [self.id, dest_id]}

        self.write_nmea(msg)

    def send_single_shot_sst(self):
        # Build the single shot CCSST message ($CCSST,1)
        msg = {'type': 'CCSST', 'params': [1]}

        self.write_nmea(msg)

    def send_fdp_ping(self, dest_id, rate=1, cdr = 0, hexdata=bytearray()):
        if self.fw_at_least("v3.0.0"):
            if cdr & 0b11000011:
                self._daemon_log.error("Invalid CDR value: {} for FDP Ping".format(cdr))
                return
            if len(hexdata) > 32:
                self._daemon_log.error("Invalid hexdata too long length: {} for FDP Ping. Max 32".format(len(hexdata)))
                return

            self.send_fdp_ping_v2(dest_id, rate, cdr, hexdata)
        else:
            self.send_fdp_ping_v1(dest_id)

    def send_fdp_ping_v2(self, dest_id, rate, cdr, hexdata=bytearray()):
        msg = {'type': 'CCCMD', 'params': ["PNG", self.id, dest_id, rate, cdr, hexstring_from_data(hexdata)]}

        self.write_nmea(msg)

    def send_fdp_ping_v1(self, dest_id):
        msg = {'type': 'CCCMD', 'params': ["PNG", self.id, dest_id, 1, 0]}

        self.write_nmea(msg)

    def wait_for_ping_reply(self, dest_id, timeout=30, include_cst=False):
        time = None
        cst = None
        ping_reply = self.wait_for_nmea_type('CAMPR', timeout=timeout)
        if include_cst:
            cst = self.wait_for_cst(timeout=1)

        if ping_reply is not None:
            dest = int(ping_reply["params"][1])
            if dest == self.id:
                time = abs(float(ping_reply["params"][2]))
        if include_cst:
            return time, cst
        else:
            return time

    def attach_incoming_ping_reply_queue(self, queue_to_attach):
        self.incoming_ping_reply_queues.append(queue_to_attach)

    def detach_incoming_ping_reply_queue(self, queue_to_detach):
        self.incoming_ping_reply_queues.remove(queue_to_detach)

    def wait_for_fdp_ping_reply(self, timeout=5, include_cst=True):
        incoming_ping_reply_queue = Queue()
        incoming_cst_queue = Queue()
        cst = None
        self.attach_incoming_ping_reply_queue(incoming_ping_reply_queue)

        try:
            ping = incoming_ping_reply_queue.get(block=True, timeout=timeout)
            if include_cst:
                self.attach_incoming_cst_queue(incoming_cst_queue)

                try:
                    cst = incoming_cst_queue.get(block=True, timeout=1)
                except Empty:
                    cst = None

                self.detach_incoming_cst_queue(incoming_cst_queue)

        except Empty:
            ping = None
            cst = None

        self.detach_incoming_ping_reply_queue(incoming_ping_reply_queue)

        return ping, cst

    def send_minipacket(self, dest_id, databytes=[]):
        msg = {'type': 'CCMUC', 'params': [self.id, dest_id, databytes[0:4]]}
        self.write_nmea(msg)

    def send_tdp(self, dest_id=None, databytes=[], rate_num=None, ack=False, base64data=False,ismini=False):
        if not ismini:
            rate = FDPDataRates[rate_num]
        else :
            rate = FDPMiniRates[rate_num]
        ack = int(ack)
        if databytes is None:
            databytes = bytearray()
            for framenum in range(0,rate.numframes):
                databytes=databytes + bytearray(range(0,rate.framesize,1))
        #databytes = bytearray(range(0, 8, 1))
        # Build the CCTDP message
        msg = {'type': 'CCTDP', 'params': [dest_id, rate_num, ack, base64data, hexstring_from_data(databytes)]}

        self.write_nmea(msg)

    def send_pgt(self, mode,txsig, rxsig1, rxsig2, rxsig3, rxsig4,bwtx,bwrx,timeout_ms):
        # Build the CCPGT message
        txc = dtcodes[txsig]
        rx1 = dtcodes[rxsig1]
        rx2 = dtcodes[rxsig2]
        rx3 = dtcodes[rxsig3]
        rx4 = dtcodes[rxsig4]
        codelen = 28
        self._daemon_log.info("sending $CCPGT with txsig,rxsig %s"% (str([txsig,rxsig1,rxsig2,rxsig3,rxsig4])))
        msg = {'type': 'CCPGT',
               'params': [mode, txc.carrierHz, codelen, txc.hexcodestr, timeout_ms, rx1.carrierHz, codelen,
                          rx1.hexcodestr, rx2.hexcodestr, rx3.hexcodestr, rx4.hexcodestr, bwtx, bwrx, 0]}
        self.write_nmea(msg)

    def send_pnt(self, tx_freq_hz, tx_duration_ms, rx_duration_ms, timeout_ms, rx_freqs_hz, sync_flag=False):
        # Build the CCPNT message
        # Pad the list of frequencies so that we have 4.  We set unused frequencies to 0
        rx_freqs_hz += [0] * (4 - len(rx_freqs_hz))

        msg = {'type': 'CCPNT',
               'params': [tx_freq_hz, tx_duration_ms, rx_duration_ms, timeout_ms,
                          rx_freqs_hz[0], rx_freqs_hz[1], rx_freqs_hz[2], rx_freqs_hz[3],
                          sync_flag]}

        self._daemon_log.info("sending $CCPNT")
        self.write_nmea(msg)

    def send_sweep(self, direction):
        '''This will send a sweep using the CCRSP command.'''
        # try to guess what was meant by the direction
        if str(direction).lower() in ['up', 'fsk', '1']:
            direction = 1
        elif str(direction).lower() in ['down', 'psk', '2']:
            direction = 2
        else:
            direction = 1

        # Build a CCRSP message
        msg = {'type': 'CCRSP', 'params': [0, direction, 0]}

        self.write_nmea(msg)

    def send_acm(self, params):
        msg = {'type': 'CCACM', 'params': params}
        self.write_nmea(msg)

    def set_config(self, name, value, response_timeout=2):
        params = [str(name), str(value)]
        msg = {'type': 'CCCFG', 'params': params}
        self.write_nmea(msg)

        if not response_timeout:
            return

        response = self.wait_for_nmea_type('CACFG', timeout=response_timeout, params=params)
        if not response:
            return None
        else:
            return {str(name), str(value)}


    def set_slot(self, value, response_timeout=2):
        #changes from recovery (0), slot 1 or slot 2
        params = [str(value)]
        msg = {'type': 'CCRST', 'params': params}
        self.write_nmea(msg)

        if not response_timeout:
            return

        response = self.wait_for_nmea_type('CACFG', timeout=response_timeout, params=params)
        if not response:
            return None
        else:
            return {str(name), str(value)}

    def ping_remus(self, group =1, channel=1, SF=0, STO = 0, response_timeout=20, AF=1, BF=1, CF=0, DF=0):
        #pings REMUS
        params = [str(group), str(channel), str(SF), str(STO), str(response_timeout), str(AF), str(BF), str(CF), str(DF)]
        msg = {'type': 'CCPDT', 'params': params}
        self.write_nmea(msg)

        response = self.wait_for_nmea_type('SNTTA', timeout=response_timeout, params=params)
        return response


        #if not response:
        #    return None
        #else:
        #    return {response}

    def ping_narrow(self, Ftx=9000, Ttx=10, Trx=10, response_timeout=10, FA=9000, FB=10000, FC=10500, FD=11000, Tflag=1):
        #pings narrowband transponder, host to modem
        params = [str(Ftx), str(Ttx), str(Trx), str(response_timeout), str(FA), str(FB), str(FC), str(FD), str(Tflag)]
        msg = {'type': 'CCPNT', 'params': params}
        self.write_nmea(msg)

        response = self.wait_for_nmea_type('SNTTA', timeout=response_timeout, params=params)
        if not response:
            return None
        else:
            return {response}

    def change_gpio(self, pin, value, toggle=False, duration=0):
        if pin == 1:
            MECValue = 5
        elif pin == 2:
            MECValue = 6
        elif pin == 3:
            MECValue = 3
        elif pin == 4:
            MECValue = 2
        elif pin == 5:
            MECValue = 4
        else:
            MECValue = 5
        if not toggle:
            if value == 0:
                mode = 1
                arg = 0
            if value == 1:
                mode = 1
                arg = 1
            if value == -1:
                mode = 0
                arg = 0
        else:
            if duration < .5:
                arg = 0
            elif duration > .5 and duration < 1:
                arg = 1
            elif duration >= 1 and duration < 2:
                arg = 2
            elif duration >= 2 and duration < 6:
                arg = 3
            elif duration >= 6 and duration < 10:
                arg = 4
            elif duration >= 10 and duration < 20:
                arg = 5
            elif duration >= 20 and duration < 30:
                arg = 6
            elif duration >= 30:
                arg = 7
            if value == 0:
                mode = 3
            else:
                mode = 3

        message = "$CCMEC,{0},{0},{1},{2},{3}".format(self.config_data["SRC"], MECValue, mode, arg)
        self.write_nmea(message)
        response = self.wait_for_nmea_type("CAMEC", timeout=self.default_nmea_timeout)
        return int(response['params'][4])

    def send_passthrough(self, msg):
        # Truncate our message to 48 characters
        info = msg[:46] + (msg[46:] and '..')
        # Build a CCRSP message
        msg = {'type': 'CCPAS', 'params': [info]}

        self.um.write_nmea(msg)

    def start_hibernate(self, wake_at=None, wake_in=None, hibernate_at=None, hibernate_in=None, disable_schedule=False,
                        ignore_response=False):
        ''' Start hibernating this modem.  This function will attempt to adapt to the limited capabilities of older
            (uM1) hardware.
        :param wake_at: Absolute time at which to wake.  May be a datetime object, Unix timestamp, or ISO 8601 datetime
            string.  If this is not None, wake_in may *not* be set.
        :param wake_in: Relative time to hibernate before waking.  May be a timedelta object, number of seconds to
            hibernate, or an ISO 8601 duration string.  If this is not None, wake_at may *not* be set.
        :param hibernate_at: Absolute time at which to start hibernate.  May be a datetime object, Unix timestamp, or
            ISO 8601 datetime string.  If this is not None, hibernate_in may *not* be set.
        :param hibernate_in: Relative time to wait before starting hibernate.  May be a timedelta object, number of
            seconds to wait, or an ISO 8601 duration string.  If this is not None, hibernate_at may *not* be set.
        :param disable_schedule: (Modem API level >= 10) If this is True and other parameters are all None, this will
            set the wake mode to 0, which causes the modem to ignore any defined wake schedule in its configuration and
            hibernate until an external wake signal is asserted.
        :param ignore_response: (ignored if API Level < 10) if this is True, this command won't block until a CAHIB
            response is received (but immediately return None).  If False (default), this function will return the
            parameters from the CAHIB message.
        '''
        # Make sure that we aren't overconstrained by parameters
        if wake_at is not None and wake_in is not None:
            raise (ValueError("Can't specify both an wake time (wake_at) and wake interval (wake_in)"))
        if hibernate_at is not None and hibernate_in is not None:
            raise (
            ValueError("Can't specify both a hibernate time (hibernate_at) and hibernate interval (hibernate_in)"))

        # Convert the parameters into more useful versions.  This will attempt to parse ISO date/duration strings.
        if wake_at is not None:
            wake_at = timeutil.convert_to_datetime(wake_at)
        if wake_in is not None:
            wake_in = timeutil.convert_to_timedelta(wake_in)
        if hibernate_at is not None:
            hibernate_at = timeutil.convert_to_datetime(hibernate_at)
        if hibernate_in is not None:
            hibernate_in = timeutil.convert_to_timedelta(hibernate_in)

        # Now, we need to generate a message, which varies depending on the API Level (uM1 vs uM2, mostly.)
        if self._api_level < 10:
            # This is a uM1, so we need to use CCMSC, with all its limitations
            # First, are we ever going to wake?
            if wake_at is None and wake_in is None:
                sleep_arg = -1
            else:
                # We need to get a duration in minutes to hibernate.
                if wake_at is not None:
                    sleep_delta = wake_at - datetime.utcnow()
                else:
                    sleep_delta = wake_in

                # We can only hibernate for multiples of 6 minutes.  Yeah...
                # First, round up to the nearest minute.
                # The following intermediate calculation is required by Python 2.6, which lacks the total_seconds() function.
                sleep_delta_seconds = sleep_delta.days * 84000 + sleep_delta.seconds
                sleep_arg = (sleep_delta_seconds + 60) // 60
                # Now round up to the nearest 6 minute interval (this takes advantage of the floor in integer division)
                sleep_arg = ((sleep_arg + 6) // 6) * 6

            if self._api_level < 6:
                # This API level is defined to be firmware that doesn't include delayed hibernate
                if (hibernate_at is not None) or (hibernate_in is not None):
                    # Not all uM1 firmware supports delayed hibernate.  The API level must be manually set to use this feature.
                    raise (
                    UnavailableInApiLevelError("This API level (and probably modem) doesn't support delayed hibernate"))

                msg = {'type': 'CCMSC', 'params': [self.id, self.id, sleep_arg]}
                self.write_nmea(msg)
                return None

            else:  # API level 6-9
                # We need to get a duration in seconds to delay hibernate.
                if hibernate_at is not None:
                    delay_delta = hibernate_at - datetime.utcnow()
                    hibernate_delay_secs = delay_delta.days * 84000 + delay_delta.seconds
                elif hibernate_in is not None:
                    hibernate_delay_secs = hibernate_in.days * 84000 + hibernate_in.seconds
                else:
                    hibernate_delay_secs = 0

                msg = {'type': 'CCMSC', 'params': [self.id, self.id, sleep_arg, hibernate_delay_secs]}
                self.write_nmea(msg)
                return None

        else:  # API level 10 and higher (uM2)
            # Make a CCHIB command.
            # Figure out the modes
            if hibernate_in is not None:
                hibernate_time = int(hibernate_in.total_seconds())
            elif hibernate_at is not None:
                hibernate_time = timeutil.to_utc_iso8601(hibernate_at, strip_fractional_seconds=True)
            else:
                hibernate_time = 0

            if wake_in is not None:
                wake_time = int(wake_in.total_seconds())
            elif wake_at is not None:
                wake_time = timeutil.to_utc_iso8601(wake_at, strip_fractional_seconds=True)
            elif not disable_schedule:
                wake_time = 0
            else:
                wake_time = 0

            msg = {'type': 'CCHIB', 'params': [hibernate_time, wake_time]}
            self.write_nmea(msg)

            if not ignore_response:
                response = self.wait_for_nmea_type("CAHIB", timeout=self.default_nmea_timeout)
                if response is None:
                    return None
                # parse the response.
                ret = namedtuple("CAHIB", ["hibernate_cause", "hibernate_time", "wake_cause", "wake_time"])
                ret.hibernate_cause = int(response['params'][0])
                ret.hibernate_time = timeutil.convert_to_datetime(response['params'][1])
                ret.wake_cause = int(response['params'][2])
                ret.wake_time = timeutil.convert_to_datetime(response['params'][3])
                return ret

            return None

    def set_host_clock_from_modem(self):
        self.set_host_clock_flag = True        
        msg = {'type':'CCCLQ', 'params':[0,0]}
        self.write_nmea(msg)
        # The actual clock setting is done by the CACLQ parser when the flag is true.

    def set_time(self, time_to_set=None, mode=None, ignore_response=False, extpps_drive_fxn=None):
        ''' Set the modem clock to the specified time (or the current time of the host system if time_to_set is not
         specified.
         Requires that the modem support the $CCTMS command, which any modern firmware does.
        '''
        if time_to_set is None:
            time_to_set = datetime.utcnow()

        if mode is None:
            mode = 0

        # Generate the command
        cmd = {'type': "CCTMS", 'params': ["{0}Z".format(timeutil.to_utc_iso8601(time_to_set, True)), mode]}
        # Send it.
        self.write_nmea(cmd)

        # If the user has specified a function to call to control the EXTPPS line in mode 1, call it.
        if (mode == 1) and (extpps_drive_fxn is not None):
            extpps_drive_fxn()

        # If timeout is not None, check the response.
        if not ignore_response:
            # CCTMS may take up to 3 seconds to time out in mode 1.
            response = self.wait_for_nmea_type('CATMS', timeout=4)
            if response is None:
                # We timed out... this could be an error condition
                return None
            # Return the argumenst of the CATMS command, more or less...
            CATMS = namedtuple("CATMS", ["time", "timed_out"])
            ret = CATMS(time=timeutil.convert_to_datetime(response['params'][1]), timed_out=response['params'][0])
            return ret
        else:
            return None

    def get_time(self, timeout=0.5):
        modem_time = self.get_time_info(timeout)[0]
        return modem_time

    def get_time_info(self, timeout=0.5):
        # For now, don't support the old $CCCLQ command.
        if self._api_level < 11:
            raise (UnavailableInApiLevelError(
                "This API level doesn't support the enhanced time-query functionality.  Use the $CCCLQ command manually."))

        self.write_nmea("$CCTMQ,0")

        response = self.wait_for_nmea_type("CATMQ", timeout=timeout)
        if response is None:
            return None

        # first argument is the time as an ISO8601 string.
        modem_time = timeutil.convert_to_datetime(response['params'][0])
        clock_source = response['params'][1]
        pps_source = response['params'][2]

        return (modem_time, clock_source, pps_source)

    def wait_for_minipacket(self, timeout=30):
        data = None
        userminipacket_recpt = self.wait_for_nmea_type('CAMUA', timeout=timeout)

        if userminipacket_recpt is not None:
            dest = int(userminipacket_recpt["params"][1])
            if dest == self.id:
                data = str(userminipacket_recpt["params"][2])
        return data

    def attach_incoming_dataframe_queue(self, queue_to_attach):
        self.incoming_dataframe_queues.append(queue_to_attach)

    def detach_incoming_dataframe_queue(self, queue_to_detach):
        self.incoming_dataframe_queues.remove(queue_to_detach)

    def attach_packetrx_queue(self, queue_to_attach):
        self.packetrx_queues.append(queue_to_attach)

    def detach_packetrx_queue(self, queue_to_detach):
        self.packetrx_queues.remove(queue_to_detach)

    def wait_for_data_packet(self, fsk=False, timeout=30):
        data_frame_queue = Queue()
        self._daemon_log.debug("wait_for_data_packet: Waiting for CARXP")
        self.wait_for_nmea_type(type_string='CARXP', timeout=timeout)
        self.attach_incoming_dataframe_queue(data_frame_queue)
        self._daemon_log.debug("wait_for_data_packet: Got CARXP")
        # If FSK packet, wait for secondary CACYC
        if fsk:
            # Wait for CST for CACYC
            self._daemon_log.debug("wait_for_data_packet: Waiting for CACYC CST")
            cst = self.wait_for_cst(timeout=timeout)
        # Wait for CST for Data packet.
        cst = self.wait_for_cst(timeout=timeout)
        self.detach_incoming_dataframe_queue(data_frame_queue)

        self._daemon_log.debug("wait_for_data_packet: Processing Data Packets Received.")
        data = bytearray()
        # Reject packet if some of the data didn't make it or if the message received wasn't for me.
        if cst is None or cst['bad_frames_num'] > 0 or cst['dest'] != self.id:
            self._daemon_log.warn("CST not valid. {}".format(cst))
            return None
        frame_count = cst['num_frames']
        self._daemon_log.debug("wait_for_data_packet: Number of Data Frames expected: {}".format(frame_count))
        while frame_count > 0:
            try:
                data_frame = data_frame_queue.get(block=True, timeout=timeout)
            except Empty:
                # Reject packet if not enough data frames were received.
                self._daemon_log.debug(
                    "wait_for_data_packet: Empty Dataframe Queue before total frames received. Frames Left:{}".format(
                        frame_count))
                return None
            # Reject data_frames not destined for me.
            if data_frame.dest != self.id:
                self._daemon_log.debug("wait_for_data_packet: Skipping data frame not destined for {}".format(self.id))
                continue
            data.extend(data_frame.data)
            self._daemon_log.debug("wait_for_data_packet: Processed Frame #{}.".format(frame_count))
            frame_count = frame_count - 1

        self._daemon_log.info("wait_for_data_packet: Returning Data ({}).".format(repr(data)))
        return data

    def attach_incoming_cst_queue(self, queue_to_attach):
        self.incoming_cst_queues.append(queue_to_attach)

    def detach_incoming_cst_queue(self, queue_to_detach):
        self.incoming_cst_queues.remove(queue_to_detach)

    def wait_for_cst(self, timeout=None):
        incoming_cst_queue = Queue()
        self.attach_incoming_cst_queue(incoming_cst_queue)

        try:
            cst = incoming_cst_queue.get(block=True, timeout=timeout)
        except Empty:
            cst = None

        self.detach_incoming_cst_queue(incoming_cst_queue)

        return cst

    def attach_incoming_xst_queue(self, queue_to_attach):
        self.incoming_xst_queues.append(queue_to_attach)

    def detach_incoming_xst_queue(self, queue_to_detach):
        self.incoming_xst_queues.remove(queue_to_detach)

    def wait_for_xst(self, timeout=None):
        incoming_xst_queue = Queue()
        self.attach_incoming_xst_queue(incoming_xst_queue)

        try:
            xst = incoming_xst_queue.get(block=True, timeout=timeout)
        except Empty:
            xst = None

        self.detach_incoming_xst_queue(incoming_xst_queue)

        return xst

    def attach_incoming_msg_queue(self, queue_to_attach):
        self.incoming_msg_queues.append(queue_to_attach)

    def detach_incoming_msg_queue(self, queue_to_detach):
        self.incoming_msg_queues.remove(queue_to_detach)

    def attach_incoming_modem_log_queue(self, queue_to_attach):
        self.incoming_log_queues.append(queue_to_attach)

    def detach_incoming_modem_log_queue(self, queue_to_detach):
        self.incoming_log_queues.remove(queue_to_detach)

    def attach_incoming_causb2_queue(self, queue_to_attach):
        self.incoming_causb2_queues.append(queue_to_attach)

    def detach_incoming_causb2_queue(self, queue_to_detach):
        self.incoming_causb2_queues.remove(queue_to_detach)

    def attach_incoming_casst_queue(self, queue_to_attach):
        self.incoming_casst_queues.append(queue_to_attach)

    def detach_incoming_casst_queue(self, queue_to_detach):
        self.incoming_casst_queues.remove(queue_to_detach)

    def wait_for_regex(self, regex_pattern, timeout=None):
        incoming_msg_queue = Queue()
        self.attach_incoming_msg_queue(incoming_msg_queue)

        regex = re.compile(regex_pattern)
        matching_msg = None

        remaining_time = timeout
        if remaining_time is not None:
            # If this program is ported to Python 3, this should be changed to use time.steady().
            end_time = time() + timeout

        while (remaining_time is None) or (remaining_time > 0):
            try:
                new_msg = incoming_msg_queue.get(timeout=remaining_time)
                matchobj = regex.search(new_msg['raw'])
                if matchobj is not None:
                    matching_msg = new_msg
                    break
                else:
                    if remaining_time is not None:
                        remaining_time = end_time - time()
                    continue
            except Empty:
                break

        self.detach_incoming_msg_queue(incoming_msg_queue)

        return matching_msg

    def wait_for_nmea_type(self, type_string, timeout=None, params=None):
        incoming_msg_queue = Queue()
        self.attach_incoming_msg_queue(incoming_msg_queue)

        matching_msg = None

        remaining_time = timeout
        if remaining_time is not None:
            # If this program is ported to Python 3, this should be changed to use time.steady().
            end_time = time() + timeout

        while (remaining_time is None) or (remaining_time > 0):
            try:
                new_msg = incoming_msg_queue.get(timeout=remaining_time)
                if new_msg['type'] == type_string:
                    if params:
                        if len(new_msg['params']) != len(params):
                            continue
                        for (n, p) in zip(new_msg['params'], params):
                            # make sure that any non-None parameter matches.
                            if p and (n != p):
                                continue

                    matching_msg = new_msg
                    break
                else:
                    if remaining_time is not None:
                        remaining_time = end_time - time()
                    continue
            except Empty:
                break

        self.detach_incoming_msg_queue(incoming_msg_queue)

        return matching_msg

    def wait_for_nmea_types(self, type_string_list, timeout=None):
        incoming_msg_queue = Queue()
        self.attach_incoming_msg_queue(incoming_msg_queue)

        matching_msg = None

        remaining_time = timeout
        if remaining_time is not None:
            # If this program is ported to Python 3, this should be changed to use time.steady().
            end_time = time() + timeout

        while (remaining_time is None) or (remaining_time > 0):
            try:
                new_msg = incoming_msg_queue.get(timeout=remaining_time)
                if new_msg['type'] in type_string_list:
                    matching_msg = new_msg
                    break
                else:
                    if remaining_time is not None:
                        remaining_time = end_time - time()
                    continue
            except Empty:
                break

        self.detach_incoming_msg_queue(incoming_msg_queue)

        return matching_msg

    def set_uplink_data_function(self, func):
        if hasattr(func, '__call__'):
            self.get_uplink_data_function = func

    """
    def request_modem_log(self, all_or_newest=1, order=0, num_to_retrieve=0, filter_params=[], timeout=None):
        filtr = BitArray(hex='0x00')
        if 'Modem To Host' in filter_params:
            filtr.set(1, 1)
        if 'Host to Modem' in filter_params:
            filtr.set(1, 2)
        if 'CARXD' in filter_params:
            filtr.set(1, 3)
        if 'CACST' in filter_params:
            filtr.set(1, 4)
        if 'CAXST' in filter_params:
            filtr.set(1, 5)

        params = [1, all_or_newest, order, num_to_retrieve, filtr.uint]

        # Build the CCRBR message
        msg = {'type': 'CCRBR', 'params': params}
        # Not this waits for a response of
        self.write_nmea(msg)

        response = self.wait_for_nmea_type('CARBR', timeout=timeout, params=[1, 0, '', '', ''])
    """

    def update_firmware(self, firmware_file_path, slot=1, reboot=False, File_location=0, data_upload_callback_fxn=None,
                        done_call_back_fxn=None):
        size_of_acceptable_data_chunks = 1013

        filesize = os.path.getsize(firmware_file_path)
        self.packets_to_send = filesize / size_of_acceptable_data_chunks
        self.packet_count = 0
        filename = os.path.basename(firmware_file_path)

        rev_timeout = self.get_config_param('CTO')

        if self.connection.can_change_baudrate():
            # Set Baud Rate on Modem
            self.set_config(name='uart1.bitrate', value=115200, response_timeout=None)
            # Change Local Baudrate
            self.connection.change_baudrate(115200)
            response = self.wait_for_nmea_type('CAREV', timeout=rev_timeout + 5)
            if response is None:
                self._daemon_log.error("REV message not received. Aborting.")
                return

        data_timeout_rate = size_of_acceptable_data_chunks / (self.connection.baudrate / 8)

        with open(firmware_file_path, 'rb') as firmware_file:
            sha1hash = hashlib.sha1(firmware_file.read(filesize)).hexdigest()
            firmware_file.seek(0)

            # Send the Update Modem FW Command
            params = [slot, File_location, filesize, int(reboot), sha1hash, filename]
            msg = {'type': 'UPMFW', 'params': params}
            self.write_nmea(msg)
            response = self.wait_for_nmea_types(['UPMFWA', 'UPERR'], timeout=2)
            if response is None:
                self._daemon_log.error("Update FW Command not acknowledged. Aborting.")
                return
            elif response['type'] == 'UPERR':
                if 'Restarting' not in response['params'][1]:
                    self._daemon_log.error("Update FW Command Error ({}). Aborting.".format(response['params'][1]))
                else:
                    self._daemon_log.info(
                        "Update FW Command Error ({}). Continuing Update.".format(response['params'][1]))
            elif response['type'] == 'UPMFWA':
                for (n, p) in zip(response['params'], params):
                    # make sure that any non-None parameter matches.
                    if p and (n != p):
                        self._daemon_log.error(
                            "Update FW Command Parameter Mismatch ({} != {} ). Aborting.".format(params,
                                                                                                 response['params']))
                        return

            # Send our firmware file.
            data_chunk = firmware_file.read(size_of_acceptable_data_chunks)
            while data_chunk != "":
                params = [hexlify(bytes(data_chunk))]
                msg = {'type': 'UPDAT', 'params': params}
                self.write_nmea(msg)
                response = self.wait_for_nmea_types(['UPDATA', 'UPERR'], timeout=data_timeout_rate + 2)
                if response is None:
                    self._daemon_log.error("Update FW Data Upload not acknowledged. Aborting.")
                    return
                elif response['type'] == 'UPERR':
                    self._daemon_log.error("Update FW Data Upload Error ({}). Aborting.".format(response['params'][1]))
                    return
                elif response['type'] == 'UPDATA':
                    if int(response['params'][0]) != len(data_chunk):
                        self._daemon_log.error(
                            "Update FW Data Upload Error ({} != {}). Aborting.".format(len(data_chunk),
                                                                                       response['params'][0]))
                        return
                    self.packet_count += 1
                    if data_upload_callback_fxn is not None:
                        data_upload_callback_fxn()
                data_chunk = firmware_file.read(size_of_acceptable_data_chunks)

            response = self.wait_for_nmea_types(['UPDONE', 'UPERR'], timeout=None)
            if response['type'] == 'UPDONE':
                self._daemon_log.info("Update FW Command Completed. ({})".format(response['params']))
                if done_call_back_fxn is not None:
                    done_call_back_fxn()
            elif response['type'] == 'UPERR':
                self._daemon_log.error("Update FW Update Error ({}). Aborting.".format(response['params'][1]))


class Message(dict):
    def __init__(self, raw):
        """Strips off NMEA checksum and leading $, returns command dictionary
        (or throws an exception if the checksum is invalid)."""
        dict.__init__(self)
        if raw == "": return None
        msg = raw.lstrip('$').rstrip('\r\n').rsplit('*')
        if len(msg) == 2:
            msg, chksum = msg
            if len(chksum) == 2:
                correctChksum = nmeaChecksum(msg)
            elif len(chksum) == 8:
                correctChksum = nmeaChecksum32(msg)
            if chksum and (chksum != correctChksum):
                # print "\nCheksum Bad. Modem: {}; Check Func: {}\n".format(chksum, correctChksum)
                raise ChecksumException("Checksum Error. Rec'd: %s, Exp'd: %s\n" % (chksum, correctChksum))

        else:
            msg = msg[0]

        msgParts = [part.strip() for part in msg.split(',')]  # splits on commas and removes spaces/CRLF
        self['type'] = msgParts[0]
        self['params'] = msgParts[1:]
        self['raw'] = raw

    def __repr__(self):
        message = ",".join([str(p) for p in [self['type']] + self['params']])
        chk = nmeaChecksum(message)
        message = "$" + message.lstrip('$').rstrip('\r\n*') + "*" + chk + "\r\n"

        return message




#!/usr/bin/env python
#
#	Run this script with ipython for interactive testing:
#   $ ipython -i test_interactive_local_acomms.py
#
#   When run this way, the interpreter runs the python code below then
#   drops you into an interactive shell to test say, um.send_single_shot_sst()

#   !! You must have your $PYTHONPATH set correctly so it has the path for 
#   !! the local pyacomms directory (e.g., /home/user/pyacomms) run:
#   $ export PYTHONPATH:/home/user/pyacomms:$PYTHONPATH

#   I recommend running in a virtual env to make sure you're not using the pip package

from acomms import micromodem, unifiedlog
import logging

unified_log = unifiedlog.UnifiedLog(log_path='./', console_log_level=logging.INFO)

um = micromodem.Micromodem(name='Micromodem2',unified_log=unified_log)

print('\n---')
print('Not currently connected to a micromodem!')
print('To connect:...... um.connect_serial(\'/dev/ttyUSB0\', 19200)')
print('To disconnect:... um.disconnect()')
print('\n---')



'''
Created on Jan 25, 2012

@author: Eric
'''

from acomms.messageparams import Packet, Rates, DataFrame

import threading


class CommState(object):
    """
    classdocs
    """

    # Default timeout for each state.  Call timeout method after this time has elapsed.
    timeout_seconds = 10

    def __init__(self, modem):
        self.modem = modem
        self.state_timer = None

    def entering(self):
        self.modem._daemon_log.debug("Entering new state: " + str(self))

        # Stop any timeout timer that is currently running.
        try:
            self.modem.state_timer.cancel()
        except:
            pass

        self.state_timer = None

        # Start a new timeout timer, if this state requires one.
        if self.timeout_seconds:
            self.state_timer = threading.Timer(self.timeout_seconds, self.timeout)

    def got_cacyc(self, cycleinfo):
        self.modem._daemon_log.debug("[" + str(self) + "]: Got CACYC")
        pass

    def got_catxf(self):
        self.modem._daemon_log.debug("[" + str(self) + "]: Got CATXF")
        pass

    def got_cadrq(self, drqparams):
        self.modem._daemon_log.debug("[" + str(self) + "]: Got CADRQ")
        pass

    def got_carx(self, rxdataframe):
        self.modem._daemon_log.debug("[" + str(self) + "]: Got CARXD")
        pass

    def got_badcrc(self):
        self.modem._daemon_log.debug("[" + str(self) + "]: Got Bad CRC")
        pass

    def got_datatimeout(self, frame_num):
        self.modem._daemon_log.debug("[" + str(self) + "]: Got DATA_TIMEOUT")
        pass

    def got_packettimeout(self):
        self.modem._daemon_log.debug("[" + str(self) + "]: Got PACKET_TIMEOUT")
        pass

    def got_pskerror(self, message):
        self.modem._daemon_log.debug("[" + str(self) + "]: Got PSK Error")
        pass

    def got_carev(self, msg):
        self.modem._daemon_log.debug("[" + str(self) + "]: Got CAREV")
        # See if we just rebooted.
        if msg['params'][1] == "AUV":
            self.modem.fw_version = msg['params'][2]
        if msg['params'][1] == "INIT":
            pass
        if self.modem.current_txpacket:
            self.modem.on_packettx_failed()
        self.modem._changestate(Idle)

    
    def got_caerr(self,hhmmss,module,err_num,message):
        self.modem._daemon_log.debug("[" + str(self) + "]: Got CAERR. (Time:%s Module: %s Number:%d Message:%s)" % (hhmmss,module,err_num,message))
        pass
    
    def got_camsg(self,msg_type,number):
        self.modem._daemon_log.debug("[" + str(self) + "]: Got CAMSG. (Type:%s Number:%d)" % (msg_type,number))
        pass
    
    def got_campa(self, src, dest):
        self.modem._daemon_log.debug("[" + str(self) + "]: Heard Ping Request From %d to %d" % (src,dest))
        pass

    def got_camea(self, src, dest):
        self.modem._daemon_log.debug("[" + str(self) + "]: Heard Line Control Command From %d to %d" % (src,dest))
        pass

    def got_camsa(self, src, dest):
        self.modem._daemon_log.debug("[" + str(self) + "]: Heard Sleep Command From %d to %d" % (src,dest))
        pass

    def got_minipacket_cmd(self, cmd, src):
        self.modem._daemon_log.debug("[{state}]: Got {cmd} Command from {src}".format(self, cmd, src))

    def send_minipacket_cmd(self, cmd, dest):
        pass

    def got_cadqf(self, dqf, packet_type):
        self.modem._daemon_log.debug("[" + str(self) + "]: Got CADQF. (DQF:%d Packet Type:%d)" % (dqf,packet_type))
        pass

    def sent_cccyc(self, cycleinfo):
        self.modem._daemon_log.debug("[" + str(self) + "]: Sent CCCYC")
        pass

    def timeout(self):
        self.modem._daemon_log.debug("[" + str(self) + "]: Timed out")
        pass

    def send_fdpacket(self, fdpacket, use_tdp=False):
        self.modem._daemon_log.warn("Trying to send packet while modem is {0}".format(self.__class__.__name__))
    
    def send_packet(self, packet):
        self.modem._daemon_log.warn("Trying to send packet while modem is {0}.".format(self.__class__.__name__))


class Idle(CommState):

    # Idle is the only state that never times out.
    timeout_seconds = None

    def entering(self):
        super(Idle, self).entering()

        # Reset our status objects
        self.modem.current_rx_frame_num = 1
        self.modem.current_tx_frame_num = 1
        self.modem.current_cycleinfo = None
        self.modem.current_txpacket = None
        self.modem.current_rxpacket = None
        
    def sent_cccyc(self, cycleinfo):
        super(Idle, self).sent_cccyc(cycleinfo)
        
        self.modem.current_cycleinfo = cycleinfo
        
        self.modem._changestate(WaitingForCacyc)
        
    def got_cacyc(self, cycleinfo):
        super(Idle, self).got_cacyc(cycleinfo)
        
        self.modem.current_cycleinfo = cycleinfo
        
        # check to see if this is an uplink request or incoming data
        if (cycleinfo.src == self.modem.id):
            self.modem._changestate(WaitingForDrq)
        else:
            self.modem._changestate(WaitingForRxData)

    def send_fdpacket(self, fdpacket, use_tdp=False):
        self.modem._daemon_log.info("Sending FD Packet from Idle")
        self.modem._daemon_log.debug("FDPacket: {}".format(dir(fdpacket)))

        if use_tdp:
            self.modem._daemon_log.debug("FDPacket using CCTDP")
            msg = fdpacket.msg_tdp()
        else:
            ver_pass = self.modem.fw_at_least('2.1.9')
            msg = fdpacket.msg_tfp(ver_pass)

        self.modem._daemon_log.debug("FDPacket: {}".format(msg))
        self.modem.write_nmea(msg)

        self.modem._changestate(WaitingForTxf)

    def send_packet(self, packet):
        self.modem._daemon_log.info("Sending packet from Idle")
        
        # copy the packet to the output buffer
        self.modem.current_txpacket = packet
        self.modem.current_cycleinfo = packet.cycleinfo
        
        # Send the cycle init
        self.modem.send_cycleinit(packet.cycleinfo)
        
        self.modem._changestate(WaitingForCacyc)

    def send_minipacket_cmd(self, cmd, dest):
        self.modem._daemon_log.info("Sending {cmd} Command from Idle".format(cmd))
        self.modem.send_ping()
        self.modem._changestate(WaitingForMinipacketTxf)

    def got_minipacket_cmd(self, cmd, src):
        super(Idle, self).got_minipacket_cmd(cmd, src)

        # Once we get a minipacket command, we send a reply...
        self.modem._changestate(WaitingForMinipacketTxf)

    def got_camsg(self,msg_type,number):
        super(Idle,self).got_camsg(msg_type,number)

    def __str__(self):
        return "Idle"


class WaitingForCacyc(CommState):
    ''' State: Waiting for the modem to echo a $CACYC message in response to a $CCCYC message that we issued. '''

    def entering(self):
        super(WaitingForCacyc, self).entering()
        
    def got_cacyc(self, cycleinfo):
        super(WaitingForCacyc, self).got_cacyc(cycleinfo)
        
        # Make sure this is the CACYC we expect
        # If not, it arrived acoustically or from another serial port, and we need to abort 
        # our current transmit to receive this packet.
        if (cycleinfo == self.modem.current_cycleinfo):
            # Is this an uplink or downlink?
            if (cycleinfo.src == self.modem.id):
                # Downlink.  Is it an FSK (rate 0) downlink (will it send a minipacket CI)?
                if cycleinfo.rate_num == 0:
                    self.modem._changestate(WaitingForCiTxf)
                else:
                    self.modem._changestate(WaitingForDrq)
            else:
                # Uplink.  Wait for the cycle init to finish transmitting.
                self.modem._changestate(WaitingForCiTxf)
        else:
            # This CACYC doesn't match what we expect
            # TODO: Add log message here
            
            # Abort the transmit
            self.modem.on_packettx_failed()
            self.modem.current_cycleinfo = cycleinfo
            
            # Is this an uplink request for us?
            if cycleinfo.src == self.modem.id:
                self.modem._changestate(WaitingForDrq)
            else:
                self.modem._changestate(WaitingForRxData)
                
        # if we got here, something is wrong
        #self.modem._daemon_log.warn("Impossible state in WaitingForCacyc.got_cacyc")
                    
    def timeout(self):
        super(WaitingForCacyc, self).timeout()
        
        # TODO: Log this error
        self.modem.on_packettx_failed()
        
        self.modem._changestate(Idle)
        
    def __str__(self):
        return "Waiting For CACYC"
        

class WaitingForDrq(CommState):
    ''' State: Waiting for modem to issue a $CADRQ message, since we got a CACYC and plan to transmit data. '''

    def entering(self):
        super(WaitingForDrq, self).entering()

    def got_cadrq(self, drqparams):
        super(WaitingForDrq, self).got_cadrq(drqparams)
        
        # See if we have a downlink packet to send
        if self.modem.current_txpacket != None:
            # Make sure that this DRQ matches the packet we plan to send
            #if 1#(drqparams.src == self.modem.current_txpacket.cycleinfo.src
            #        and drqparams.dest == self.modem.current_txpacket.cycleinfo.dest
            #        and drqparams.num_bytes == Rates[self.modem.current_txpacket.cycleinfo.rate_num].framesize
            #        and drqparams.frame_num == self.modem.current_tx_frame_num):
                # Send the frame
            self.modem._send_current_txframe()
                
                # Was this the last frame?
            if drqparams.frame_num < self.modem.current_txpacket.cycleinfo.num_frames:
                    # It wasn't, so get ready to send the next frame.
                self.modem.current_tx_frame_num += 1
                self.modem._changestate(WaitingForDrq)
            else:
                    # That was the last frame, so wait for the packet to transmit
                self.modem._changestate(WaitingForTxf)
                    
            #else:
                # The DRQ we just received does not match the packet we have queued for transmit.
                # Therefore, we don't know what state we're in.  Say that we failed and go back to Idle.
            #    self.modem.on_packettx_failed()
            #    self.modem._changestate(Idle)
        else:
            # We don't have a current TX packet, so this is an uplink request
            
            # This is probably not always what we want to do
            #TODO: Add better uplink support
            self.modem.send_uplink_frame(drqparams)
            
            # Do we expect to send more frames?
            if drqparams.frame_num < self.modem.current_cycleinfo.num_frames:
                self.modem._changestate(WaitingForDrq)
            else:
                self.modem._changestate(WaitingForTxf)
    
    def got_datatimeout(self, frame_num):
        super(WaitingForDrq, self).got_datatimeout(frame_num)
        
        # Note that if we are doing an uplink and we are configured not to respond, this is expected behavior.
        
        # See if we are trying to send a packet.
        if self.modem.current_txpacket != None:
            # We failed.  Oh well.
            self.modem.on_packettx_failed()
            
        # Now, see if the modem will still transmit for any reason
        if self.modem.asd == True:
            # The modem will generate a test frame and still transmit
            self.modem._changestate(WaitingForTxf)
        else:
            # See if we already sent a frame, in which case the modem will still transmit.
            if frame_num > 1:
                self.modem._changestate(WaitingForTxf)
            else:
                self.modem._changestate(Idle)
    
    def timeout(self):
        super(WaitingForDrq, self).timeout()
        
        # See if we are trying to send a packet
        if self.modem.current_txpacket != None:
            # We failed.  Signal an error
            self.modem.on_packettx_failed()
        
        # We don't really know what state we're in.
        self.modem._changestate(Idle)
        
    def got_catxf(self):
        super(WaitingForDrq, self).got_catxf()
        
        # We can get here if there was a frame in the mdoem's FIFO from an ealier CCTXD message
        # However, we don't know what that frame might contain.  It's probably not what we want to send.
        if self.modem.current_txpacket != None:
            self.modem.on_packettx_failed()
            
        self.modem._changestate(Idle)
        
    def __str__(self):
        return "Waiting for CADRQ"
    

class WaitingForTxf(CommState):
    ''' State: Waiting for the $CATXF message when we expect the modem to transmit. '''

    # TODO: Adjust this interval based on the packet type, number of frames, and BW0.
    timeout_seconds = 60

    def entering(self):
        super(WaitingForTxf, self).entering()

    def got_catxf(self):
        super(WaitingForTxf, self).got_catxf()
        
        # If we were trying to send a packet, we just succeeded.
        if self.modem.current_txpacket != None:
            self.modem.on_packettx_success()
        
        # Now, go back to Idle.
        self.modem._changestate(Idle)

    def timeout(self):
        super(WaitingForTxf, self).timeout()

        # See if we are trying to send a packet
        if self.modem.current_txpacket != None:
            # We failed.  Signal an error
            self.modem.on_packettx_failed()

        # We don't really know what state we're in.
        self.modem._changestate(Idle)
        
    def __str__(self):
        return "Waiting for CATXF"


class WaitingForCiTxf(CommState):
    ''' State: Waiting for the $CATXF message when we expect the modem to transmit a cycle init. '''
    timeout_seconds = 30

    def entering(self):
        super(WaitingForCiTxf, self).entering()
        
    def got_catxf(self):
        super(WaitingForCiTxf, self).got_catxf()
        
        # See if we are in the middle of an FSK downlink
        if self.modem.current_txpacket != None:
            self.modem._changestate(WaitingForDrq)
        else:
            self.modem._changestate(WaitingForPacket)
    
    def timeout(self):
        super(WaitingForCiTxf, self).timeout()
        
        # We timed out, so signal an error
        # Note that the modem might not really be done transmitting.
        
        # Were we in the middle of an FSK downlink?
        if self.modem.current_txpacket != None:
            self.modem.on_packettx_failed()
        else:
            self.modem.on_uplink_failed()
        
        self.modem._changestate(Idle)
        
    def __str__(self):
        return "Waiting for CI CATXF"


class WaitingForMinipacketTxf(CommState):
    def got_catxf(self):
        super(WaitingForMinipacketTxf, self).got_catxf()

        # That should do it.
        self.modem._changestate(Idle)

    def timeout(self):
        super(WaitingForMinipacketTxf, self).timeout()

        # We timed out, so signal an error
        # Note that the modem might not really be done transmitting.
        self.modem.on_minipacket_tx_failed()

        self.modem._changestate(Idle)

    def __str__(self):
        return "Waiting for Minipacket Command CATXF"


class WaitingForPacket(CommState):
    ''' State: Waiting for a packet following an uplink request. '''
    timeout_seconds = 120

    def entering(self):
        super(WaitingForPacket, self).entering()

    def got_cacyc(self, cycleinfo):
        super(WaitingForPacket, self).got_cacyc(cycleinfo)
        
        # Make sure this is the packet we expect.
        if cycleinfo == self.modem.current_cycleinfo:
            # This is the packet for which we were waiting, so wait for the CARXDs.
            self.modem._changestate(WaitingForRxData)
        else:
            # This was not the CACYC we expected.
            # Replace the current_cycleinfo to match this incoming message and raise an error
            self.modem.on_uplink_failed()
            self.modem.current_cycleinfo = cycleinfo
            
            self.modem._changestate(WaitingForRxData)
        
    def got_packettimeout(self):
        super(WaitingForPacket, self).got_packettimeout()
        
        # TODO: Raise error here
        
        self.modem._changestate(Idle)
        
    def timeout(self):
        super(WaitingForPacket, self).timeout()
        
        #TODO: Raise error here
        
        self.modem._changestate(Idle)
        
    def __str__(self):
        return "Waiting for packet"


class WaitingForRxData(CommState):
    ''' State: Waiting for $CARXD or $CARXA messages after getting a $CACYC message that suggests we should.  '''

    def entering(self):
        super(WaitingForRxData, self).entering()
        
        self.modem.current_rxpacket = Packet(cycleinfo=self.modem.current_cycleinfo)
        
    def got_carx(self, rxdataframe):
        super(WaitingForRxData, self).got_carx(rxdataframe)
        
        # Make sure that the frame we received has parameters matching the packet we think we're receiving.
        if (rxdataframe.src == self.modem.current_rxpacket.cycleinfo.src
                and rxdataframe.dest == self.modem.current_rxpacket.cycleinfo.dest) or self.modem._ignore_cacyc_address_mismatch:
            # Copy this new data to our packet in progress
            self.modem.current_rxpacket.frames.append(rxdataframe)

            # Check to see if this frame had the ACK bit set.
            if rxdataframe.ack:
                self.modem.current_rxpacket.cycleinfo.ack = True
            
            # See if this was the last frame
            if rxdataframe.frame_num >= self.modem.current_rxpacket.cycleinfo.num_frames:
                # We got a complete packet
                # See if it had BAD_CRCs on any frame.
                if any(frame.bad_crc for frame in self.modem.current_rxpacket.frames):
                    self.modem.on_packetrx_failed
                else:
                    self.modem.on_packetrx_success(self.modem.current_rxpacket)

                # Are we sending ACKs?  If so, wait for that.
                if self.modem.current_rxpacket.cycleinfo.ack:
                    self.modem._changestate(WaitingForMinipacketTxf)
                else:
                    # Go back to Idle
                    self.modem._changestate(Idle)
            

            else:
                # We are waiting for another frame
                pass
        else:
            # This is not the frame we were expecting
            # We can't make a good packet, so signal an error
            self.modem.on_packetrx_failed()
            self.modem._changestate(Idle)
        
    def got_badcrc(self):
        super(WaitingForRxData, self).got_badcrc()

        # Add this BAD_CRC as a "frame" to the current packet.
        this_frame_num = self.modem.current_rxpacket.frames[-1].frame_num + 1
        this_quote_frame = DataFrame(self.modem.current_rxpacket.cycleinfo.src,
                                     self.modem.current_rxpacket.cycleinfo.dest,
                                     self.modem.current_rxpacket.cycleinfo.ack,
                                     this_frame_num,
                                     None,
                                     bad_crc=True)
        self.modem.current_rxpacket.frames.append(this_quote_frame)

        # See if this was the last frame
        if this_frame_num >= self.modem.current_rxpacket.cycleinfo.num_frames:
            # This packet is done, and it failed.
            self.modem.on_packetrx_failed()
            self.modem._changestate(Idle)
        else:
            # We are still waiting for the next frame.
            pass

    def got_packettimeout(self):
        super(WaitingForRxData, self).got_packettimeout()

        # TODO: Raise error here

        self.modem._changestate(Idle)
        
    def timeout(self):
        super(WaitingForRxData, self).timeout()
        
        # If we are receiving a packet, we just failed.
        if self.modem.current_rxpacket != None:
            self.modem.on_packetrx_failed()
        
        self.modem._changestate(Idle)

    def got_camsg(self,msg_type,number):
        super(WaitingForRxData, self).got_camsg(msg_type, number)
        self.modem.on_packetrx_failed()
        self.modem._changestate(Idle)

    def __str__(self):
        return "Waiting for RX Data"



__author__ = 'Eric'

import logging
import os
import time
from datetime import datetime


class UnifiedLog(object):
    def __init__(self, log_path=None, file_name=None, console_log_level=None, rootname=None, custom_handler=None):

        # Create the logger
        if rootname is not None:
            self._log = logging.getLogger(rootname)
        else:
            self._log = logging.getLogger()

        # Use UTC timestamps in ISO8601 format
        self._logformat = logging.Formatter("%(asctime)s\t%(levelname)s\t%(name)s\t%(message)s", "%Y-%m-%dT%H:%M:%SZ")
        self._logformat.converter = time.gmtime

        if console_log_level is not None:
            self._console_handler = logging.StreamHandler()
            self._console_handler.setLevel(console_log_level)
            self._console_handler.setFormatter(self._logformat)
            self._log.addHandler(self._console_handler)

        if custom_handler is not None:
            self._log.addHandler(custom_handler)

        # If no log path is specified, use (or create) a directory in the user's home directory
        if log_path is None:
            log_path = os.path.expanduser('~/acomms_logs')

        log_path = os.path.normpath(log_path)

        # Create the directory if it doesn't exist
        if not os.path.isdir(log_path):
            os.makedirs(log_path)

        if file_name is None:
            now = datetime.utcnow()
            file_name = "acomms_{0}.log".format(now.strftime("%Y%m%dT%H%M%SZ"))

        log_file_path = os.path.join(log_path, file_name)

        self.log_file_name = log_file_path

        self._file_handler = logging.FileHandler(log_file_path)
        self._file_handler.setLevel(logging.DEBUG)
        self._file_handler.setFormatter(self._logformat)
        self._log.addHandler(self._file_handler)

    def log_file_path(self): #Added to try and help error in downlinktest_all.py -Dana
        return self.log_file_name

    def getLogger(self, name):
        return self._log.getChild(name)

    def start_django_output(self, model, test_script_run):
        self._django_handler = DjangoHandler(model, test_script_run)
        self._log.addHandler(self._django_handler)


class DjangoHandler(logging.Handler):
    def __init__(self, model=None, test_script_run=None):
        super(DjangoHandler,self).__init__()
        self.model = model
        self.test_script_run = test_script_run

    def emit(self, record):
        ts = datetime.utcfromtimestamp(record.created)
        entry = self.model(test_script_run=self.test_script_run, timestamp=ts, level=record.levelname, name=record.name, message=record.message)
        entry.save()



from binascii import hexlify, unhexlify
from builtins import bytes

def data_from_hexstring(hexstring):
    databytes = bytes.fromhex(hexstring)
    return databytes
    """
    databytes = bytearray()
    try:
        databytes.extend([ord(c) for c in unhexlify(hexstring)])
    #Catch Odd-length String Error
    except TypeError:
        pass

    return databytes"""


def hexstring_from_data(databytes):
    hex_databytes = hexlify(databytes).decode("utf-8")
    return hex_databytes


class CycleInfo(object):
    def __init__(self, src, dest, rate_num, ack=False, num_frames=None):
        self.src = int(src)
        self.dest = int(dest)
        self.rate_num = int(rate_num)
        self.ack = bool(ack)

        if num_frames == None:
            self.num_frames = Rates[rate_num].numframes
        else:
            self.num_frames = int(num_frames)

    # This allows us to see if two cycleinfo objects match
    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class DrqParams(object):
    def __repr__(self):
        return "SRC: {} DST: {} ACK: {} NUMBYTES: {} FRAME#: {}".format(self.src, self.dest, self.ack, self.num_bytes,
                                                                        self.frame_num)

    def __init__(self, src, dest, ack, num_bytes, frame_num):
        self.src = int(src)
        self.dest = int(dest)
        self.ack = bool(ack)
        self.num_bytes = int(num_bytes)
        self.frame_num = int(frame_num)


class Ack(object):
    def __repr__(self):
        return "SRC: {} DST: {} FRAME#: {}".format(self.src, self.dest, self.frame_num)

    def __init__(self, src, dest, ack, frame_num):
        self.src = int(src)
        self.dest = int(dest)
        self.ack = bool(ack)
        self.frame_num = int(frame_num)


class Camua(object):
    def __init__(self, src, dest, data):
        self.src = src
        self.dest = dest
        self.data = bytearray(data)

    def __repr__(self):
        return "SRC: {} DST: {} DATA: {}".format(self.src, self.dest, repr(self.data))


class Campr(object):
    def __init__(self, src, dest, owtt):
        self.src = src
        self.dest = dest
        self.owtt = owtt

class Causb2(object):
    def __init__(self, df, iso8601_time, azimuth_deg, elevation_deg, owtt_s, timing_mode,
                 ire0, ire1, ire2, ire3,
                 ndetects):

        self.df = df
        self.iso8601_time = iso8601_time
        self.azimuth_deg = azimuth_deg
        self.elevation_deg = elevation_deg
        self.owtt_s = owtt_s
        self.timing_mode = timing_mode

        self.ire0 = ire0
        self.ire1 = ire1
        self.ire2 = ire2
        self.ire3 = ire3

        self.ndetects = ndetects

class Casst(object):
    def __init__(self, sst_version, in_water_spl_dB, detector, num_samples, summary):
        self.sst_version = sst_version
        self.in_water_spl_dB = in_water_spl_dB
        self.detector = detector
        self.num_samples = num_samples

        self.summary = summary

    def __doc__(self, *args, **kwargs):
        return '''
        SST-- SoundSTatistics            
            -Print descriptive statistics about the in-water Sound Pressure Level (SPL).  
        
        How it works: 
        -------------
        Observations are accumulated beginning after the previous CASST was printed.  These samples of SPL data are stored in a ring buffer with enough memory to hold 250-ish observations which is roughly 10 seconds-worth (or the classic CAREV period) given a detector probe length of 200 symbols.  The buffer of observations is cleared after every time it is printed; the final field in the SST message reports the number of samples that were accumulated.  The SPL samples are tied to a specific detector and config; each SPL observation is the average SPL, in dB, over the detector's probe length (in baseband samples, something like 40 mS).
        

        MicroModem HW & SW configuration that affects CASST output:
        -----------------------------------------------------------
        ***Currently only works for PSK detectors operating at 10 kHz or 25 kHz (i.e. BND 1, BND 3 and BND 0 **with FC0 configured properly)
        Detector:
            -Must be using a PSK detector, o/wise, all 0's in output
        Carrier:
            -Only tuned for 25 kHz and 10 kHz, anything else will return 0's for SPL 
        Bandwidth:
            -This determines your passband for the SPL estimate 
        FML:
            -Sweep length - i.e. window size for averaging; modifies number of baseband samples in observation
        AGN:
            -Analog gain setting for the detector (modem's front end) 
        Hardware:
            -hwd.Vpreamp_gain_dB: transducer's preamp gain, measured from ducer to modem by convention. default is 40 dB for preamp matched to BTech 25 kHz ducer. 
            -hwd.transducer_sensitivity_dB: transducer sensitivity in dB, measured from wavefront in water to preamp by convention. default is -193 dB for BTech 25 kHz ducer. 


        Single Shot, Prints $CASST immediatly: 
        --------------------------------------
            -$CCSST,1 -------------------- um.send_single_shot_sst()

        Continuous Printing (prints with each CAREV):  
        --------------------------------------
            -To enable: $CCCFG,SST,1 ----- um.set_config(type='SST', 1)
            -To disable: $CCCFG,SST,0 ---- um.set_config(type='SST', 0)
            -To query: $CCCFQ,SST -------- um.get_config(type='SST')


        $CASST,0,58.728850,1,400,57.862410,58.381225,58.523375,58.728850,59.203695,191,*70
               |   |       |  |    |         |         |         |         |        |______ number of elements in SPL history.  Each element is the average of a unique set of nObservationSamples of baseband data
               |   |       |  |    |         |         |         |         |_______________ sample maximum of SPL history (in dB).  Max of five-number summary (descriptive statistics)               
               |   |       |  |    |         |         |         |_________________________ upper quartile of SPL history (in dB).  Third quartile of five-number summary (descriptive statistics) 
               |   |       |  |    |         |         |___________________________________ median value of SPL history (in dB).  Middle value of five-number summary (descriptive statistics) 
               |   |       |  |    |         |_____________________________________________ lower quartile of SPL history (in dB).  First quartile of five-number summary (descriptive statistics)     
               |   |       |  |    |_______________________________________________________ sample minimum of SPL history (in dB).  Min of five-number summary (descriptive statistics)            
               |   |       |  |____________________________________________________________ nObservationSamples: number of baseband samples in the detector's probe
               |   |       |_______________________________________________________________ detector used during SPL calculation. This enum represents the detector type (e.g., psk chan, fsk chan, etc...) *only supports psk chan as of 3.2.2
               |   |_______________________________________________________________________ in-water Sound Pressure Level (SPL) in dB. This is the average SPL over the last nObservationSamples (see above for nObservationSamples definition)
               |___________________________________________________________________________ SST version                   
        '''

    def __repr__(self, *args, **kwargs):
        return "SST_VER: {} IN_WATER_SPL_DB: {} DETECTOR: {} #SAMPLES: {} SUMMARY: {}".format(
            self.sst_version,
            self.in_water_spl_dB,
            self.detector,
            self.num_samples,
            self.summary,
            )


class DataFrame(object):
    def __init__(self, src, dest, ack, frame_num, data, bad_crc=False):
        self.bad_crc = bad_crc
        self.src = src
        self.dest = dest
        self.ack = ack
        self.frame_num = frame_num
        if data:
            self.data = bytearray(data)
        else:
            self.data = None

    def __repr__(self):
        return "SRC: {} DST: {} ACK: {} FRAME#: {} BAD_CRC: {} DATA: {}".format(self.src, self.dest, self.ack,
                                                                                self.frame_num, self.bad_crc,
                                                                                repr(self.data))


class CCPGT(object):
    def __init__(self, txfreq, txcode, timeout_ms, codelen, rxfreq, rxcode1, rxcode2, rxcode3, rxcode4, reserved1=0,
                 reserved2=0):
        self.txfreq = txfreq
        self.codelen = codelen
        self.txcode = txcode
        self.timeout_ms = timeout_ms
        self.rxfreq = rxfreq
        self.rxcode1 = rxcode1
        self.rxcode2 = rxcode2
        self.rxcode3 = rxcode3
        self.rxcode4 = rxcode4
        self.reserved1 = reserved1
        self.reserved2 = reserved2

        def __repr__(self):
            return "TXHz: {} RXHz: {} TXCODE: {}".format(self.txfreq, self.rxfreq, self.txcode)


class Packet(object):
    def __init__(self, cycleinfo, frames=None):
        self.cycleinfo = cycleinfo

        if frames != None:
            self.frames = frames
        else:
            self.frames = []

    def append_framedata(self, framedata):
        # TODO: Make sure we have room for another frame, and that the data fits in the frame.
        newframe = DataFrame(self.cycleinfo.src, self.cycleinfo.dest, self.cycleinfo.ack,
                             (len(self.frames) + 1), framedata)
        self.frames.append(newframe)


class PacketRate(object):
    def __init__(self, name, number, framesize, numframes):
        self.name = name
        self.number = number
        self.framesize = framesize
        self.numframes = numframes

    def getpacketsize(self):
        return self.framesize * self.numframes

    maxpacketsize = property(getpacketsize)


Rates = {0: PacketRate('FH-FSK', 0, 32, 1),
         1: PacketRate('BCH 128:8', 1, 64, 3),
         2: PacketRate('DSS 1/15 (64B frames)', 2, 64, 3),
         3: PacketRate('DSS 1/7', 3, 256, 2),
         4: PacketRate('BCH 64:10', 4, 256, 2),
         5: PacketRate('Hamming 14:9', 5, 256, 8),
         6: PacketRate('DSS 1/15 (32B frames)', 6, 32, 6)}

FDPMiniRates = {1: PacketRate('BCH 128:8', 1, 90, 1),
                3: PacketRate('BCH 64:10', 3, 90, 1),
                5: PacketRate('Hamming 14:9', 5, 90, 1)}
FDPDataRates = {1: PacketRate('BCH 128:8', 1, 64, 3),
                3: PacketRate('BCH 64:10', 3, 256, 2),
                5: PacketRate('Hamming 14:9', 5, 256, 8)}
LDRRates = {7: PacketRate('BCH 64:10', 1, 260, 1)}



from acomms import Micromodem
from acomms.flexibledataprotocol import FDPacket, FDFrame

from acomms.net_packet import NetPacket

class NetModem(Micromodem):
    """Network enabled micromodem driver implementation"""

    def __init__(self, name='netmodem', unified_log=None,
                 log_path=None, log_level='INFO', sound_speed=1500, max_range_m = 5000):
        super(NetModem, self).__init__(name=name, unified_log=unified_log,
                                       log_path=log_path, log_level=log_level)

        # Use sound speed for calculating ranges and timeouts
        self.packet_timeout = max_range_m/sound_speed

        # Keep track of active links by the destination SRC ID
        # self.link_manager = LinkManager()

        # Acoustic networking is implemented on top of micromodem
        # Flexible data packets
        self.fdp_listeners.append(self.on_fdp_receive)

        # Different network protocols have their own handlers
        self.protocol_handlers = {}
        self.protocol_queues = {}

    def register_protocol_queue(self, pid, queue):
        self.protocol_queues.setdefault(pid, []).append(queue)

    def register_protocol_handler(self, pid, handler_func):
        self.protocol_handlers.setdefault(pid, []).append(handler_func)

    def on_fdp_receive(self, fdpacket):
        netpacket = NetPacket.from_fdpacket(fdpacket)

        callbackset = set()
        queueset = set()

        # Check each of the subheaders in the packet
        for key in netpacket.subheaders:
            subheader = netpacket.subheaders[key]
            pid = subheader[0]

            # For each PID with a registered handler, pass handler
            # the subheader and the netpacket
            if pid in self.protocol_handlers:
                for func in self.protocol_handlers[pid]:
                    callbackset.add(func)

            # For each PID with a registered queue, put
            # the netpacket on the queue
            if pid in self.protocol_queues:
                for queue in self.protocol_queues[pid]:
                    queueset.add(queue)

        # Callback all the functions in the set
        for func in callbackset:
            try:
                func(netpacket)
            except Exception as e:
                self._daemon_log.warn("Error with PID handler function: " + str(func) + " " + str(e))

        # Append netpacket to all queues in set
        for queue in queueset:
            try:
                queue.put_nowait(netpacket)
            except:
                self._daemon_log.warn("Error appending netpacket to protocol queue")

    def transmit_netpacket(self, netpacket, miniframerate, dataframerate, ack):
        self.send_fdpacket(netpacket.to_fdpacket(miniframerate=miniframerate,
                                                 dataframerate=dataframerate,
                                                 ack=ack))



__author__ = 'Eric Gallimore'


from acomms import micromodem, unifiedlog
import logging
from time import sleep
import argparse

if __name__ == '__main__':
        ap = argparse.ArgumentParser(description ='Forward some CST data from one modem acoustically using another modem')
        ap.add_argument("logpath", help="Location of Log File", default="/home/acomms/")
        ap.add_argument("-cs","--cstserial", help='CST modem serial port', default="/dev/ttyO1")
        ap.add_argument("-cb","--cstbaud", help="CST modem Baud Rate", default=19200)
        ap.add_argument("-ts","--txserial", help='TX modem serial port', default="/dev/ttyO2")
        ap.add_argument("-tb","--txbaud", help="TX modem baud rate", default=19200)
        ap.add_argument("-d","--dest", help="Destination modem address", default=1)


        args = ap.parse_args()

        unified_log = unifiedlog.UnifiedLog(log_path=args.logpath, console_log_level=logging.INFO)

        cst_modem = micromodem.Micromodem(name='CST_modem',unified_log=unified_log)
        tx_modem = micromodem.Micromodem(name="TX_modem", unified_log=unified_log)

        cst_modem.connect_serial(args.cstserial, args.cstbaud)
        tx_modem.connect_serial(args.txserial, args.txbaud)

        try:
            while True:
                new_cst = cst_modem.wait_for_cst()

                tx_modem.send_packet_data(args.dest, bytes(new_cst.snr_in));

        finally:
            cst_modem.disconnect()
            tx_modem.disconnect()